import torch
from torch import sin, cos
from torch import autograd
from .neurodiffeq import safe_diff as diff


def _split_u_x(*us_xs):
    if len(us_xs) == 0 or len(us_xs) % 2 != 0:
        raise RuntimeError("Number of us and xs must be equal and positive")
    us = us_xs[:len(us_xs) // 2]
    xs = us_xs[len(us_xs) // 2:]
    return us, xs


def grad(u, *xs):
    r"""Gradient of tensor u with respect to a tuple of tensors xs.
    Given :math:`u` and :math:`x_1`, ..., :math:`x_n`, the function returns
    :math:`\frac{\partial u}{\partial x_1}`, ..., :math:`\frac{\partial u}{\partial x_n}`

    :param u: The :math:`u` described above.
    :type u: `torch.Tensor`
    :param xs: The sequence of :math:`x_i` described above.
    :type xs: `torch.Tensor`
    :return: A tuple of :math:`\frac{\partial u}{\partial x_1}`, ..., :math:`\frac{\partial u}{\partial x_n}`
    :rtype: List[`torch.Tensor`]
    """
    grads = []
    for x, g in zip(xs, autograd.grad(u, xs, grad_outputs=torch.ones_like(u), create_graph=True, allow_unused=True)):
        if g is None:
            grads.append(torch.zeros_like(x, requires_grad=True))
        else:
            grads.append(g.requires_grad_(True))
    return grads


def div(*us_xs):
    r"""Derives and evaluates the divergence of a :math:`n`-dimensional vector field :math:`\mathbf{u}`
    with respect to :math:`\mathbf{x}`.

    :param us_xs:
        The input must have :math:`2n` tensors, each of shape (n_samples, 1)
        with the former :math:`n` tensors being the entries of :math:`u`
        and the latter :math:`n` tensors being the entries of :math:`x`.
    :type us_xs: `torch.Tensor`
    :return: The divergence evaluated at :math:`x`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    us, xs = _split_u_x(*us_xs)
    return sum(diff(u, x) for u, x in zip(us, xs))


def curl(u_x, u_y, u_z, x, y, z):
    r"""Derives and evaluates the curl of a vector field :math:`\mathbf{u}` in three dimensional cartesian coordinates.

    :param u_x: The :math:`x`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_x: `torch.Tensor`
    :param u_y: The :math:`y`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_y: `torch.Tensor`
    :param u_z: The :math:`z`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_z: `torch.Tensor`
    :param x: A vector of :math:`x`-coordinate values, must have shape (n_samples, 1).
    :type x: `torch.Tensor`
    :param y: A vector of :math:`y`-coordinate values, must have shape (n_samples, 1).
    :type y: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return: The :math:`x`, :math:`y`, and :math:`z` components of the curl, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """
    dxy, dxz = grad(u_x, y, z)
    dyx, dyz = grad(u_y, x, z)
    dzx, dzy = grad(u_z, x, y)

    return dzy - dyz, dxz - dzx, dyx - dxy


def laplacian(u, *xs):
    r"""Derives and evaluates the laplacian of a scalar field :math:`u`
    with respect to :math:`\mathbf{x}=[x_1, x_2, \dots]`

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param xs: The sequence of :math:`x_i` described above. Each with shape (n_samples, 1)
    :type xs: `torch.Tensor`
    :return: The laplacian of :math:`u` evaluated at :math:`\mathbf{x}`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    gs = grad(u, *xs)
    return sum(diff(g, x) for g, x in zip(gs, xs))


def vector_laplacian(u_x, u_y, u_z, x, y, z):
    r"""Derives and evaluates the vector laplacian of a vector field :math:`\mathbf{u}`
    in three dimensional cartesian coordinates.

    :param u_x: The :math:`x`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_x: `torch.Tensor`
    :param u_y: The :math:`y`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_y: `torch.Tensor`
    :param u_z: The :math:`z`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_z: `torch.Tensor`
    :param x: A vector of :math:`x`-coordinate values, must have shape (n_samples, 1).
    :type x: `torch.Tensor`
    :param y: A vector of :math:`y`-coordinate values, must have shape (n_samples, 1).
    :type y: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return:
        Components of vector laplacian of :math:`\mathbf{u}` evaluated at :math:`\mathbf{x}`,
        each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """
    return laplacian(u_x, x, y, z), laplacian(u_y, x, y, z), laplacian(u_z, x, y, z)


def spherical_curl(u_r, u_theta, u_phi, r, theta, phi):
    r"""Derives and evaluates the spherical curl of a spherical vector field :math:`u`.

    :param u_r: The :math:`r`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_r: `torch.Tensor`
    :param u_theta: The :math:`\theta`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_theta: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The :math:`r`, :math:`\theta`, and :math:`\phi` components of the curl, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """

    ur_dth, ur_dph = grad(u_r, theta, phi)
    uth_dr, uth_dph = grad(u_theta, r, phi)
    uph_dr, uph_dth = grad(u_phi, r, theta)
    csc_th = 1 / sin(theta)
    r_inv = 1 / r

    curl_r = r_inv * (uph_dth + (u_phi * cos(theta) - uth_dph) * csc_th)
    curl_th = r_inv * (csc_th * ur_dph - u_phi) - uph_dr
    curl_ph = uth_dr + r_inv * (u_theta - ur_dth)

    return curl_r, curl_th, curl_ph


def spherical_grad(u, r, theta, phi):
    r"""Derives and evaluates the spherical gradient of a spherical scalar field :math:`u`.

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The :math:`r`, :math:`\theta`, and :math:`\phi` components of the gradient, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """
    u_dr, u_dth, u_dph = grad(u, r, theta, phi)
    r_inv = 1 / r
    return u_dr, u_dth * r_inv, u_dph * r_inv / sin(theta)


def spherical_div(u_r, u_theta, u_phi, r, theta, phi):
    r"""Derives and evaluates the spherical divergence of a spherical vector field :math:`u`.

    :param u_r: The :math:`r`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_r: `torch.Tensor`
    :param u_theta: The :math:`\theta`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_theta: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The divergence evaluated at :math:`(r, \theta, \phi)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    sin_th = sin(theta)
    return (diff(u_r * r ** 2, r) / r + (diff(u_theta * sin_th, theta) + diff(u_phi, phi)) / sin_th) / r


def spherical_laplacian(u, r, theta, phi):
    r"""Derives and evaluates the spherical laplacian of a spherical scalar field :math:`u`.

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The laplacian evaluated at :math:`(r, \theta, \phi)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    u_dr, u_dth, u_dph = grad(u, r, theta, phi)
    sin_th = sin(theta)
    r2 = r ** 2

    return (diff(r2 * u_dr, r) + diff(sin_th * u_dth, theta) / sin_th + diff(u_dph, phi) / sin_th ** 2) / r2


def spherical_vector_laplacian(u_r, u_theta, u_phi, r, theta, phi):
    r"""Derives and evaluates the spherical laplacian of a spherical vector field :math:`u`.

    :param u_r: The :math:`r`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_r: `torch.Tensor`
    :param u_theta: The :math:`\theta`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_theta: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param r: A vector of :math:`r`-coordinate values, must have shape (n_samples, 1).
    :type r: `torch.Tensor`
    :param theta: A vector of :math:`\theta`-coordinate values, must have shape (n_samples, 1).
    :type theta: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :return: The laplacian evaluated at :math:`(r, \theta, \phi)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    ur_dr, ur_dth, ur_dph = grad(u_r, r, theta, phi)
    uth_dr, uth_dth, uth_dph = grad(u_theta, r, theta, phi)
    uph_dr, uph_dth, uph_dph = grad(u_phi, r, theta, phi)
    sin_th, cos_th = sin(theta), cos(theta)
    sin2_th = sin_th ** 2
    r2 = r ** 2

    scalar_lap_r = (diff(r2 * ur_dr, r) + diff(sin_th * ur_dth, theta) / sin_th + diff(ur_dph, phi) / sin2_th) / r2
    scalar_lap_th = (diff(r2 * uth_dr, r) + diff(sin_th * uth_dth, theta) / sin_th + diff(uth_dph, phi) / sin2_th) / r2
    scalar_lap_ph = (diff(r2 * uph_dr, r) + diff(sin_th * uph_dth, theta) / sin_th + diff(uph_dph, phi) / sin2_th) / r2

    vec_lap_r = scalar_lap_r - 2 * (u_r + uth_dth + (cos_th * u_theta + uph_dph) / sin_th) / r2
    vec_lap_th = scalar_lap_th + (2 * ur_dth - (u_theta + 2 * cos_th * uph_dph) / sin2_th) / r2
    vec_lap_ph = scalar_lap_ph + ((2 * cos_th * uth_dph - u_phi) / sin_th + 2 * ur_dph) / (r2 * sin_th)

    return vec_lap_r, vec_lap_th, vec_lap_ph


def spherical_to_cartesian(r, theta, phi):
    r"""Convert spherical coordinate :math:`(r, \theta, \phi)` to cartesian coordinates :math:`(x, y, z)`.
    The input shapes of r, theta, and phi must be the same.

    :param r: The :math:`r`-component of spherical coordinates.
    :type r: `torch.Tensor`
    :param theta: The :math:`\theta`-component (polar angle) of spherical coordinates.
    :type theta: `torch.Tensor`
    :param phi: The :math:`\phi`-component (azimuthal angle) of spherical coordinates.
    :type phi: `torch.Tensor`
    :return: The :math:`x`-, :math:`y`-, and :math:`z`-component in cartesian coordinates.
    :rtype: tuple[`torch.Tensor`]
    """
    rho = r * sin(theta)
    return rho * cos(phi), rho * sin(phi), r * cos(theta)


def cartesian_to_spherical(x, y, z):
    r"""Convert cartesian coordinates :math:`(x, y, z)` to spherical coordinate :math:`(r, \theta, \phi)`.
    The input shapes of x, y, and z must be the same.
    If either the polar angle :math:`\theta` or the azimuthal angle :math:`phi` is not defined,
    the default value will be 0.

    :param x: The :math:`x`-component of cartesian coordinates.
    :type x: `torch.Tensor`
    :param y: The :math:`y`-component of cartesian coordinates.
    :type y: `torch.Tensor`
    :param z: The :math:`z`-component of cartesian coordinates.
    :type z: `torch.Tensor`
    :return: The :math:`r`-, :math:`\theta`-, and :math:`\phi`-component in spherical coordinates.
    :rtype: tuple[`torch.Tensor`]
    """
    rho2 = x ** 2 + y ** 2
    return torch.sqrt(rho2 + z ** 2), torch.atan2(torch.sqrt(rho2), z), torch.atan2(y, x)


def cylindrical_grad(u, rho, phi, z):
    r"""Derives and evaluates the cylindrical gradient of a cylindrical scalar field :math:`u`.

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param rho: A vector of :math:`\rho`-coordinate values, must have shape (n_samples, 1).
    :type rho: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return: The :math:`\rho`, :math:`\phi`, and :math:`z` components of the gradient, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """
    u_drho, u_dphi, u_dz = grad(u, rho, phi, z)
    return u_drho, u_dphi / rho, u_dz


def cylindrical_div(u_rho, u_phi, u_z, rho, phi, z):
    r"""Derives and evaluates the cylindrical divergence of a cylindrical vector field :math:`u`.

    :param u_rho: The :math:`\rho`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_rho: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param u_z: The :math:`z`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_z: `torch.Tensor`
    :param rho: A vector of :math:`\rho`-coordinate values, must have shape (n_samples, 1).
    :type rho: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return: The divergence evaluated at :math:`(\rho, \phi, z)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    return diff(u_rho, rho) + (u_rho + diff(u_phi, phi)) / rho + diff(u_z, z)


def cylindrical_curl(u_rho, u_phi, u_z, rho, phi, z):
    r"""Derives and evaluates the cylindrical curl of a cylindrical vector field :math:`u`.

    :param u_rho: The :math:`\rho`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_rho: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param u_z: The :math:`z`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_z: `torch.Tensor`
    :param rho: A vector of :math:`\rho`-coordinate values, must have shape (n_samples, 1).
    :type rho: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return: The :math:`\rho`, :math:`\phi`, and :math:`z` components of the curl, each with shape (n_samples, 1).
    :rtype: tuple[`torch.Tensor`]
    """
    urho_dphi, urho_dz = grad(u_rho, phi, z)
    uphi_drho, uphi_dz = grad(u_phi, rho, z)
    uz_drho, uz_dphi = grad(u_z, rho, phi)

    return (
        uz_dphi / rho - uphi_dz,
        urho_dz - uz_drho,
        uphi_drho + (u_phi - urho_dphi) / rho
    )


def cylindrical_laplacian(u, rho, phi, z):
    r"""Derives and evaluates the cylindrical laplacian of a cylindrical scalar field :math:`u`.

    :param u: A scalar field :math:`u`, must have shape (n_samples, 1).
    :type u: `torch.Tensor`
    :param rho: A vector of :math:`\rho`-coordinate values, must have shape (n_samples, 1).
    :type rho: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return: The laplacian evaluated at :math:`(\rho, \phi, z)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    u_drho, u_dphi, u_dz = grad(u, rho, phi, z)
    return diff(u_drho, rho) + u_drho / rho + diff(u_dphi, phi) / rho ** 2 + diff(u_dz, z)


def cylindrical_vector_laplacian(u_rho, u_phi, u_z, rho, phi, z):
    r"""Derives and evaluates the cylindrical laplacian of a cylindrical vector field :math:`u`.

    :param u_rho: The :math:`\rho`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_rho: `torch.Tensor`
    :param u_phi: The :math:`\phi`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_phi: `torch.Tensor`
    :param u_z: The :math:`z`-component of the vector field :math:`u`, must have shape (n_samples, 1).
    :type u_z: `torch.Tensor`
    :param rho: A vector of :math:`\rho`-coordinate values, must have shape (n_samples, 1).
    :type rho: `torch.Tensor`
    :param phi: A vector of :math:`\phi`-coordinate values, must have shape (n_samples, 1).
    :type phi: `torch.Tensor`
    :param z: A vector of :math:`z`-coordinate values, must have shape (n_samples, 1).
    :type z: `torch.Tensor`
    :return: The laplacian evaluated at :math:`(\rho, \phi, z)`, with shape (n_samples, 1).
    :rtype: `torch.Tensor`
    """
    rho2 = rho ** 2
    urho_drho, urho_dphi, urho_dz = grad(u_rho, rho, phi, z)
    uphi_drho, uphi_dphi, uphi_dz = grad(u_phi, rho, phi, z)
    uz_drho, uz_dphi, uz_dz = grad(u_z, rho, phi, z)

    scalar_lap_rho = diff(urho_drho, rho) + urho_drho / rho + diff(urho_dphi, phi) / rho ** 2 + diff(urho_dz, z)
    scalar_lap_phi = diff(uphi_drho, rho) + uphi_drho / rho + diff(uphi_dphi, phi) / rho ** 2 + diff(uphi_dz, z)
    scalar_lap_z = diff(uz_drho, rho) + uz_drho / rho + diff(uz_dphi, phi) / rho ** 2 + diff(uz_dz, z)

    return (
        scalar_lap_rho - (u_rho + 2 * uphi_dphi) / rho2,
        scalar_lap_phi + (2 * urho_dphi - u_phi) / rho2,
        scalar_lap_z,
    )


def cylindrical_to_cartesian(rho, phi, z):
    r"""Convert cylindrical coordinate :math:`(\rho, \phi, z)` to cartesian coordinates :math:`(x, y, z)`.
    The input shapes of rho, phi, and z must be the same.

    :param rho: The :math:`\rho`-component of cylindrical coordinates.
    :type rho: `torch.Tensor`
    :param phi: The :math:`\phi`-component (azimuthal angle) of cylindrical coordinates.
    :type phi: `torch.Tensor`
    :param z: The :math:`z`-component of cylindrical coordinates.
    :type z: `torch.Tensor`
    :return: The :math:`x`-, :math:`y`-, and :math:`z`-component in cartesian coordinates.
    :rtype: tuple[`torch.Tensor`]
    """
    return rho * cos(phi), rho * sin(phi), z


def cartesian_to_cylindrical(x, y, z):
    r"""Convert cartesian coordinates :math:`(x, y, z)` to cylindrical coordinate :math:`(\rho, \phi, z)`.
    The input shapes of x, y, and z must be the same.
    If the azimuthal angle :math:`phi` is undefined, the default value will be 0.

    :param x: The :math:`x`-component of cartesian coordinates.
    :type x: `torch.Tensor`
    :param y: The :math:`y`-component of cartesian coordinates.
    :type y: `torch.Tensor`
    :param z: The :math:`z`-component of cartesian coordinates.
    :type z: `torch.Tensor`
    :return: The :math:`\rho`-, :math:`\phi`-, and :math:`z`-component in cylindrical coordinates.
    :rtype: tuple[`torch.Tensor`]
    """
    return torch.sqrt(x ** 2 + y ** 2), torch.atan2(y, x), z



import sys
import pytest
from pytest import raises, warns, deprecated_call
import numpy as np
import torch
# atomic-ish generator classes
from neurodiffeq.generators import Generator1D
from neurodiffeq.generators import Generator2D
from neurodiffeq.generators import Generator3D
from neurodiffeq.generators import GeneratorND
from neurodiffeq.generators import GeneratorSpherical
# complex generator classes
from neurodiffeq.generators import ConcatGenerator
from neurodiffeq.generators import StaticGenerator
from neurodiffeq.generators import PredefinedGenerator
from neurodiffeq.generators import TransformGenerator
from neurodiffeq.generators import EnsembleGenerator
from neurodiffeq.generators import FilterGenerator
from neurodiffeq.generators import ResampleGenerator
from neurodiffeq.generators import BatchGenerator
from neurodiffeq.generators import SamplerGenerator
from neurodiffeq.generators import MeshGenerator


@pytest.fixture(autouse=True)
def magic():
    MAGIC = 42
    torch.manual_seed(MAGIC)
    np.random.seed(MAGIC)
    return MAGIC


def _check_shape_and_grad(generator, target_size, *xs):
    if target_size is not None:
        if target_size != generator.size:
            print(f"size mismatch {target_size} != {generator.size}", file=sys.stderr)
            return False
    for x in xs:
        if x.shape != (generator.size,):
            print(f"Bad shape: {x.shape} != {generator.size}", file=sys.stderr)
            return False
        if not x.requires_grad:
            print(f"Doesn't require grad: {x}", file=sys.stderr)
            return False
    return True


def _check_shape_and_grad_for_N(generator, target_size, xs):
    if target_size is not None:
        if target_size != generator.size:
            print(f"size mismatch {target_size} != {generator.size}", file=sys.stderr)
            return False
    for x in xs:
        if x.shape != (generator.size,):
            print(f"Bad shape: {x.shape} != {generator.size}", file=sys.stderr)
            return False
        if not x.requires_grad:
            print(f"Doesn't require grad: {x}", file=sys.stderr)
            return False
    return True


def _check_boundary(xs, xs_min, xs_max):
    for x, x_min, x_max in zip(xs, xs_min, xs_max):
        if x_min is not None and (x < x_min).any():
            print(f"Lower than minimum: {x} <= {x_min}", file=sys.stderr)
            return False
        if x_max is not None and (x > x_max).any():
            print(f"Higher than maximum: {x} >= {x_max}", file=sys.stderr)
            return False
    return True


def _check_iterable_equal(x, y, eps=1e-5):
    for a, b in zip(x, y):
        if abs(float(a) - float(b)) > eps:
            print(f"Different values: {a} != {b}", file=sys.stderr)
            return False
    return True


def test_generator1d():
    size = 32
    generator = Generator1D(size=size, t_min=0.0, t_max=2.0, method='uniform')
    x = generator.getter()
    assert _check_shape_and_grad(generator, size, x)

    generator = Generator1D(size=size, t_min=0.0, t_max=2.0, method='equally-spaced')
    x = generator.getter()
    assert _check_shape_and_grad(generator, size, x)

    generator = Generator1D(size=size, t_min=0.0, t_max=2.0, method='equally-spaced-noisy')
    x = generator.getter()
    assert _check_shape_and_grad(generator, size, x)

    generator = Generator1D(size=size, t_min=np.log10(0.1), t_max=np.log10(2.0), method='log-spaced')
    x = generator.getter()
    assert _check_shape_and_grad(generator, size, x)

    generator = Generator1D(size=size, t_min=np.log10(0.1), t_max=np.log10(2.0), method='log-spaced-noisy')
    x = generator.getter()
    assert _check_shape_and_grad(generator, size, x)

    generator = Generator1D(size=size, t_min=np.log10(0.1), t_max=np.log10(2.0), method='log-spaced-noisy',
                            noise_std=0.01)
    x = generator.getter()
    assert _check_shape_and_grad(generator, size, x)

    with raises(ValueError):
        generator = Generator1D(size=size, t_min=0.0, t_max=2.0, method='magic')

    str(generator)
    repr(generator)


def test_generator2d():
    grid = (5, 6)
    size = grid[0] * grid[1]
    x_min, x_max = 0.0, 1.0
    y_min, y_max = -1.0, 0.0
    x_std, y_std = 0.05, 0.06

    generator = Generator2D(grid=grid, xy_min=(x_min, y_min), xy_max=(x_max, y_max), method='equally-spaced-noisy')
    x, y = generator.getter()
    assert _check_shape_and_grad(generator, size, x, y)

    generator = Generator2D(grid=grid, xy_min=(x_min, y_min), xy_max=(x_max, y_max), method='equally-spaced-noisy',
                            xy_noise_std=(x_std, y_std))
    x, y = generator.getter()
    assert _check_shape_and_grad(generator, size, x, y)

    generator = Generator2D(grid=grid, xy_min=(x_min, y_min), xy_max=(x_max, y_max), method='equally-spaced')
    x, y = generator.getter()
    assert _check_shape_and_grad(generator, size, x, y)
    assert _check_boundary((x, y), (x_min, y_min), (x_max, y_max))

    str(generator)
    repr(generator)


def test_generator3d():
    grid = (5, 6, 7)
    size = grid[0] * grid[1] * grid[2]
    x_min, x_max = 0.0, 1.0
    y_min, y_max = 1.0, 2.0
    z_min, z_max = 2.0, 3.0

    generator = Generator3D(grid=grid, xyz_min=(x_min, y_min, z_min), xyz_max=(x_max, y_max, z_max),
                            method='equally-spaced-noisy')
    x, y, z = generator.getter()
    assert _check_shape_and_grad(generator, size, x, y, z)

    generator = Generator3D(grid=grid, xyz_min=(x_min, y_min, z_min), xyz_max=(x_max, y_max, z_max),
                            method='equally-spaced')
    x, y, z = generator.getter()
    assert _check_shape_and_grad(generator, size, x, y, z)
    assert _check_boundary((x, y, z), (x_min, y_min, z_min), (x_max, y_max, z_max))

    str(generator)
    repr(generator)


def test_generatorNd():
    grid = (5,)
    r_min = (0.1,)
    r_max = (1.0,)
    r_noise_std = (0.05,)
    max_dim = 4

    while len(grid) < (max_dim + 1):
        size = np.prod(grid)

        methods = ['uniform' for m in range(len(grid))]

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=False)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())
        assert _check_boundary(generator.getter(), r_min, r_max)

        methods = ['equally-spaced' for m in range(len(grid))]

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=True)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=True, r_noise_std=r_noise_std)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=False)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())
        assert _check_boundary(generator.getter(), r_min, r_max)

        methods = ['log-spaced' for m in range(len(grid))]

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=True)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=False)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())
        assert _check_boundary(generator.getter(), r_min, r_max)

        methods = ['exp-spaced' for m in range(len(grid))]

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=True)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())

        generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                methods=methods, noisy=False)
        assert _check_shape_and_grad_for_N(generator, size, generator.getter())
        # No check_boundary because doing 10 ** x and then log10() does not always return exactly x

        if len(grid) == 3:
            methods = ['uniform', 'equally-spaced', 'log-spaced']

            generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                    methods=methods, noisy=True)
            assert _check_shape_and_grad_for_N(generator, size, generator.getter())

            generator = GeneratorND(grid=grid, r_min=r_min, r_max=r_max,
                                    methods=methods, noisy=False)
            assert _check_shape_and_grad_for_N(generator, size, generator.getter())
            assert _check_boundary(generator.getter(), r_min, r_max)

        grid += (grid[-1] + 1,)
        r_min += (r_min[-1] + 1,)
        r_max += (r_max[-1] + 1,)
        r_noise_std += (r_noise_std[-1] + 0.01,)

    str(generator)
    repr(generator)


def test_generator_spherical():
    size = 64
    r_min, r_max = 0.0, 1.0

    generator = GeneratorSpherical(size, r_min=r_min, r_max=r_max, method='equally-spaced-noisy')
    r, theta, phi = generator.get_examples()
    assert _check_shape_and_grad(generator, size, r, theta, phi)
    assert _check_boundary((r, theta, phi), (r_min, 0.0, 0.0), (r_max, np.pi, np.pi * 2))

    generator = GeneratorSpherical(size, r_min=r_min, r_max=r_max, method='equally-radius-noisy')
    r, theta, phi = generator.get_examples()
    assert _check_shape_and_grad(generator, size, r, theta, phi)
    assert _check_boundary((r, theta, phi), (r_min, 0.0, 0.0), (r_max, np.pi, np.pi * 2))

    with pytest.raises(ValueError):
        _ = GeneratorSpherical(64, method='bad_generator')

    with pytest.raises(ValueError):
        _ = GeneratorSpherical(64, r_min=-1.0)

    with pytest.raises(ValueError):
        _ = GeneratorSpherical(64, r_min=1.0, r_max=0.0)

    str(generator)
    repr(generator)


def test_concat_generator():
    size1, size2 = 10, 20
    t_min, t_max = 0.5, 1.5
    generator1 = Generator1D(size1, t_min=t_min, t_max=t_max)
    generator2 = Generator1D(size2, t_min=t_min, t_max=t_max)
    concat_generator = ConcatGenerator(generator1, generator2)
    x = concat_generator.get_examples()
    assert _check_shape_and_grad(concat_generator, size1 + size2, x)

    grid1 = (4, 4, 4)
    size1, size2, size3 = grid1[0] * grid1[1] * grid1[2], 100, 200
    generator1 = Generator3D(grid=grid1)
    generator2 = GeneratorSpherical(size2)
    generator3 = GeneratorSpherical(size3)
    concat_generator = ConcatGenerator(generator1, generator2, generator3)
    r, theta, phi = concat_generator.get_examples()
    assert _check_shape_and_grad(concat_generator, size1 + size2 + size3, r, theta, phi)

    added_generator = generator1 + generator2 + generator3
    r, theta, phi = added_generator.get_examples()
    assert _check_shape_and_grad(added_generator, size1 + size2 + size3, r, theta, phi)

    str(concat_generator)
    repr(concat_generator)


def test_static_generator():
    size = 100
    generator = Generator1D(size)
    static_generator = StaticGenerator(generator)
    x1 = static_generator.get_examples()
    x2 = static_generator.get_examples()
    assert _check_shape_and_grad(generator, size)
    assert _check_shape_and_grad(static_generator, size, x1, x2)
    assert (x1 == x2).all()

    size = 100
    generator = GeneratorSpherical(size)
    static_generator = StaticGenerator(generator)
    r1, theta1, phi1 = static_generator.get_examples()
    r2, theta2, phi2 = static_generator.get_examples()
    assert _check_shape_and_grad(generator, size)
    assert _check_shape_and_grad(static_generator, size, r1, theta1, phi1, r2, theta2, phi2)
    assert (r1 == r2).all() and (theta1 == theta2).all() and (phi1 == phi2).all()

    str(static_generator)
    repr(static_generator)


def test_predefined_generator():
    size = 100

    old_x = torch.arange(size, dtype=torch.float, requires_grad=False)
    predefined_generator = PredefinedGenerator(old_x)
    x = predefined_generator.get_examples()
    assert _check_shape_and_grad(predefined_generator, size, x)
    assert _check_iterable_equal(old_x, x)

    old_x = torch.arange(size, dtype=torch.float, requires_grad=False)
    old_y = torch.arange(size, dtype=torch.float, requires_grad=True)
    old_z = torch.arange(size, dtype=torch.float, requires_grad=False)
    predefined_generator = PredefinedGenerator(old_x, old_y, old_z)
    x, y, z = predefined_generator.get_examples()
    assert _check_shape_and_grad(predefined_generator, size, x, y, z)
    assert _check_iterable_equal(old_x, x)
    assert _check_iterable_equal(old_y, y)
    assert _check_iterable_equal(old_z, z)

    x_list = [i * 2.0 for i in range(size)]
    y_tuple = tuple([i * 3.0 for i in range(size)])
    z_array = np.array([i * 4.0 for i in range(size)]).reshape(-1, 1)
    w_tensor = torch.arange(size, dtype=torch.float)
    predefined_generator = PredefinedGenerator(x_list, y_tuple, z_array, w_tensor)
    x, y, z, w = predefined_generator.get_examples()
    assert _check_shape_and_grad(predefined_generator, size, x, y, z, w)
    assert _check_iterable_equal(x_list, x)
    assert _check_iterable_equal(y_tuple, y)
    assert _check_iterable_equal(z_array, z)
    assert _check_iterable_equal(w_tensor, w)

    str(predefined_generator)
    repr(predefined_generator)


def test_transform_generator():
    size = 100
    x = np.arange(0, size, dtype=np.float32)
    x_expected = np.sin(x)
    generator = PredefinedGenerator(x)
    transform_generator = TransformGenerator(generator, [torch.sin])
    x = transform_generator.get_examples()
    assert _check_shape_and_grad(transform_generator, size, x)
    assert _check_iterable_equal(x, x_expected)

    x = np.arange(0, size, dtype=np.float32)
    y = np.arange(0, size, dtype=np.float32)
    z = np.arange(0, size, dtype=np.float32)
    x_expected = np.sin(x)
    y_expected = y
    z_expected = -z
    generator = PredefinedGenerator(x, y, z)
    transform_generator = TransformGenerator(generator, [torch.sin, None, lambda a: -a])
    x, y, z = transform_generator.get_examples()
    assert _check_shape_and_grad(transform_generator, size, x, y, z)
    assert _check_iterable_equal(x, x_expected)
    assert _check_iterable_equal(y, y_expected)
    assert _check_iterable_equal(z, z_expected)

    transform_generator = TransformGenerator(generator, transform=lambda x, y, z: (torch.sin(x), y, -z))
    x, y, z = transform_generator.get_examples()
    assert _check_shape_and_grad(transform_generator, size, x, y, z)
    assert _check_iterable_equal(x, x_expected)
    assert _check_iterable_equal(y, y_expected)
    assert _check_iterable_equal(z, z_expected)

    str(transform_generator)
    repr(transform_generator)


def test_ensemble_generator():
    size = 100

    generator1 = Generator1D(size)
    ensemble_generator = EnsembleGenerator(generator1)
    x = ensemble_generator.get_examples()
    assert _check_shape_and_grad(ensemble_generator, size, x)

    old_x = torch.rand(size)
    old_y = torch.rand(size)
    old_z = torch.rand(size)
    generator1 = PredefinedGenerator(old_x)
    generator2 = PredefinedGenerator(old_y)
    generator3 = PredefinedGenerator(old_z)
    ensemble_generator = EnsembleGenerator(generator1, generator2, generator3)
    x, y, z = ensemble_generator.get_examples()
    assert _check_shape_and_grad(ensemble_generator, size, x, y, z)
    assert _check_iterable_equal(old_x, x)
    assert _check_iterable_equal(old_y, y)
    assert _check_iterable_equal(old_z, z)

    old_x = torch.rand(size)
    old_y = torch.rand(size)
    generator1 = PredefinedGenerator(old_x)
    generator2 = PredefinedGenerator(old_y)
    product_generator = generator1 * generator2
    x, y = product_generator.get_examples()
    assert _check_shape_and_grad(product_generator, size, x, y)
    assert _check_iterable_equal(old_x, x)
    assert _check_iterable_equal(old_y, y)

    str(ensemble_generator)
    repr(ensemble_generator)


def test_mesh_generator():
    size = 10
    x_min, x_max = 0.0, 1.0
    y_min, y_max = 1.0, 2.0
    z_min, z_max = 2.0, 3.0

    generator1 = Generator1D(size)
    mesh_generator = MeshGenerator(generator1)
    x = mesh_generator.get_examples()
    assert _check_shape_and_grad(mesh_generator, size, x)

    mesh_generator = MeshGenerator(generator1, generator1, generator1)
    m1, m2, m3 = mesh_generator.get_examples()
    assert _check_shape_and_grad(mesh_generator, (size ** 3), m1, m2, m3)

    generator1x = Generator1D(size, x_min, x_max, method="equally-spaced")
    generator1y = Generator1D(size, y_min, y_max, method="equally-spaced")
    generator1z = Generator1D(size, z_min, z_max, method="equally-spaced")
    generator3 = Generator3D((size, size, size), (x_min, y_min, z_min), (x_max, y_max, z_max), method="equally-spaced")
    mesh_generator = MeshGenerator(generator1x, generator1y, generator1z)
    m1, m2, m3 = mesh_generator.get_examples()
    g1, g2, g3 = generator3.get_examples()
    assert _check_shape_and_grad(mesh_generator, (size ** 3), m1, m2, m3)
    assert _check_iterable_equal(g1, m1)
    assert _check_iterable_equal(g2, m2)
    assert _check_iterable_equal(g3, m3)

    generator1x = Generator1D(size, x_min, x_max, method="equally-spaced")
    generator1y = Generator1D(size, y_min, y_max, method="equally-spaced")
    generator1z = Generator1D(size, z_min, z_max, method="equally-spaced")
    generator3 = Generator3D((size, size, size), (x_min, y_min, z_min), (x_max, y_max, z_max), method="equally-spaced")
    xor_generator = generator1x ^ generator1y ^ generator1z
    m1, m2, m3 = xor_generator.get_examples()
    g1, g2, g3 = generator3.get_examples()
    assert _check_shape_and_grad(xor_generator, (size ** 3), m1, m2, m3)
    assert _check_iterable_equal(g1, m1)
    assert _check_iterable_equal(g2, m2)
    assert _check_iterable_equal(g3, m3)

    str(mesh_generator)
    repr(mesh_generator)


def test_filter_generator():
    grid = (10, 10)
    size = 100

    x = [i * 1.0 for i in range(size)]
    filter_fn = lambda a: (a[0] % 2 == 0)
    filter_fn_2 = lambda a: (a % 2 == 0)
    x_expected = filter(filter_fn_2, x)

    generator = PredefinedGenerator(x)
    filter_generator = FilterGenerator(generator, filter_fn=filter_fn, update_size=True)
    x = filter_generator.get_examples()
    assert _check_shape_and_grad(filter_generator, size // 2, x)
    assert _check_iterable_equal(x_expected, x)

    x = [i * 1.0 for i in range(size)]
    y = [-i * 1.0 for i in range(size)]
    filter_fn = lambda ab: (ab[0] % 2 == 0) & (ab[1] > -size / 2)
    x_expected, y_expected = list(zip(*filter(filter_fn, zip(x, y))))
    generator = PredefinedGenerator(x, y)
    filter_generator = FilterGenerator(generator, filter_fn)
    x, y = filter_generator.get_examples()
    assert _check_shape_and_grad(filter_generator, size // 4, x, y)
    assert _check_iterable_equal(x_expected, x)
    assert _check_iterable_equal(y_expected, y)

    generator = Generator2D(grid)
    filter_fn = lambda ab: (ab[0] > 0.5) & (ab[1] < 0.5)
    filter_generator = FilterGenerator(generator, filter_fn)
    for _ in range(5):
        x, y = filter_generator.get_examples()
        assert _check_shape_and_grad(filter_generator, None, x, y)

    fixed_size = 42
    filter_generator = FilterGenerator(generator, filter_fn, size=fixed_size, update_size=False)
    for _ in range(5):
        assert _check_shape_and_grad(filter_generator, fixed_size)
        filter_generator.get_examples()

    str(filter_generator)
    repr(filter_generator)


def test_resample_generator():
    size = 100

    sample_size = size
    x_expected = np.arange(size, dtype=np.float32)
    generator = PredefinedGenerator(x_expected)
    resample_generator = ResampleGenerator(generator, size=sample_size, replacement=False)
    x = resample_generator.get_examples()
    assert _check_shape_and_grad(resample_generator, sample_size, x)
    # noinspection PyTypeChecker
    assert _check_iterable_equal(torch.sort(x)[0], x_expected)

    sample_size = size // 2
    x = np.arange(size, dtype=np.float32)
    y = np.arange(size, size * 2, dtype=np.float32)
    generator = PredefinedGenerator(x, y)
    resample_generator = ResampleGenerator(generator, size=sample_size, replacement=False)
    x, y = resample_generator.get_examples()
    assert _check_shape_and_grad(resample_generator, sample_size, x, y)
    assert _check_iterable_equal(x + 100, y)
    assert len(torch.unique(x.detach())) == len(x)

    sample_size = size * 3 // 4
    x = np.arange(size, dtype=np.float32)
    y = np.arange(size, size * 2, dtype=np.float32)
    generator = PredefinedGenerator(x, y)
    resample_generator = ResampleGenerator(generator, size=sample_size, replacement=True)
    x, y = resample_generator.get_examples()
    assert _check_shape_and_grad(resample_generator, sample_size, x, y)
    assert _check_iterable_equal(x + 100, y)
    assert len(torch.unique(x.detach())) < len(x)

    sample_size = size * 2
    x = np.arange(size, dtype=np.float32)
    y = np.arange(size, size * 2, dtype=np.float32)
    z = np.arange(size * 2, size * 3, dtype=np.float32)
    generator = PredefinedGenerator(x, y, z)
    resample_generator = ResampleGenerator(generator, size=sample_size, replacement=True)
    x, y, z = resample_generator.get_examples()
    assert _check_shape_and_grad(resample_generator, sample_size, x, y, z)
    assert _check_iterable_equal(x + 100, y)
    assert _check_iterable_equal(y + 100, z)
    assert len(torch.unique(x.detach())) < len(x)

    str(resample_generator)
    repr(resample_generator)


def test_batch_generator():
    size = 10
    batch_size = 3
    x = np.arange(size, dtype=np.float32)
    answer_x = np.arange(batch_size) % size
    generator = PredefinedGenerator(x)
    batch_generator = BatchGenerator(generator, batch_size)
    for _ in range(50):
        x = batch_generator.get_examples()
        assert _check_shape_and_grad(batch_generator, batch_size, x)
        assert _check_iterable_equal(answer_x, x)
        assert len(batch_generator.cached_xs) <= size + max(size, batch_size)
        # update answer for next iteration
        answer_x = (answer_x + batch_size) % size

    size = 3
    batch_size = 10
    x = np.arange(size, dtype=np.float32)
    y = np.arange(size, dtype=np.float32)
    answer_x = np.arange(batch_size) % size
    answer_y = np.arange(batch_size) % size
    generator = PredefinedGenerator(x, y)
    batch_generator = BatchGenerator(generator, batch_size)
    for _ in range(50):
        x, y = batch_generator.get_examples()
        assert _check_shape_and_grad(batch_generator, batch_size, x, y)
        assert _check_iterable_equal(answer_x, x)
        assert _check_iterable_equal(answer_y, y)
        assert len(batch_generator.cached_xs) <= size + max(size, batch_size)
        # update answer for next iteration
        answer_x = (answer_x + batch_size) % size
        answer_y = (answer_y + batch_size) % size

    str(batch_generator)
    repr(batch_generator)


def test_sampler_geneartor():
    size = 10
    generator = Generator1D(size)
    sampler_generator = SamplerGenerator(generator)
    x = sampler_generator.get_examples()
    assert isinstance(x, (list, tuple))
    x, = x
    assert x.shape == (size, 1)

    str(sampler_generator)
    repr(sampler_generator)


def test_legacy_module():
    with warns(FutureWarning):
        import neurodiffeq.generator

    from neurodiffeq.ode import ExampleGenerator
    from neurodiffeq.pde import ExampleGenerator2D, PredefinedExampleGenerator2D
    from neurodiffeq.pde_spherical import ExampleGenerator3D, ExampleGeneratorSpherical
    with warns(FutureWarning):
        ExampleGenerator(100)
    with warns(FutureWarning):
        ExampleGenerator2D()
    with warns(FutureWarning):
        x = torch.rand(10)
        y = torch.rand(10)
        PredefinedExampleGenerator2D(x, y)
    with warns(FutureWarning):
        ExampleGenerator3D()
    with warns(FutureWarning):
        ExampleGeneratorSpherical(100)



import torch
import warnings
import torch.nn as nn
import inspect
from inspect import signature
from abc import ABC, abstractmethod
from itertools import chain
from copy import deepcopy
from torch.optim import Adam
from .solvers_utils import PretrainedSolver
from .networks import FCNN
from ._version_utils import deprecated_alias
from .generators import GeneratorSpherical
from .generators import SamplerGenerator
from .generators import Generator1D
from .generators import Generator2D
from .generators import GeneratorND
from .function_basis import RealSphericalHarmonics
from .conditions import BaseCondition
from .neurodiffeq import safe_diff as diff
from .losses import _losses


def _requires_closure(optimizer):
    return inspect.signature(optimizer.step).parameters.get('closure').default == inspect._empty


class BaseSolver(ABC, PretrainedSolver):
    r"""A class for solving ODE/PDE systems.

    :param diff_eqs:
        The differential equation system to solve, which maps a tuple of coordinates to a tuple of ODE/PDE residuals.
        Both the coordinates and ODE/PDE residuals must have shape (-1, 1).
    :type diff_eqs: callable
    :param conditions:
        List of boundary conditions for each target function.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    :param nets:
        List of neural networks for parameterized solution.
        If provided, length must equal that of conditions.
    :type nets: list[`torch.nn.Module`], optional
    :param train_generator:
        A generator for sampling training points.
        It must provide a `.get_examples()` method and a `.size` field.
    :type train_generator: `neurodiffeq.generators.BaseGenerator`, required
    :param valid_generator:
        A generator for sampling validation points.
        It must provide a `.get_examples()` method and a `.size` field.
    :type valid_generator: `neurodiffeq.generators.BaseGenerator`, required
    :param analytic_solutions:
        **[DEPRECATED]** Pass ``metrics`` instead.
        The analytical solutions to be compared with neural net solutions.
        It maps a tuple of coordinates to a tuple of function values.
        The output shape should match that of networks.
    :type analytic_solutions: callable, optional
    :param optimizer:
        The optimizer to be used for training.
    :type optimizer: `torch.nn.optim.Optimizer`, optional
    :param criterion:
        The loss function used for training.

        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
        - If any other callable, it must map
          A) a residual tensor (shape `(n_points, n_equations)`),
          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
          so that backpropagation can be performed.

    :type criterion:
        str or `torch.nn.moduesl.loss._Loss` or callable
    :param n_batches_train:
        Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
        Defaults to 1.
    :type n_batches_train: int, optional
    :param n_batches_valid:
        Number of batches to validate in every epoch, where batch-size equals ``valid_generator.size``.
        Defaults to 4.
    :type n_batches_valid: int, optional
    :param metrics:
        Additional metrics to be logged (besides loss). ``metrics`` should be a dict where

        - Keys are metric names (e.g. 'analytic_mse');
        - Values are functions (callables) that computes the metric value.
          These functions must accept the same input as the differential equation ``diff_eq``.

    :type metrics: dict, optional
    :param n_input_units:
        Number of input units for each neural network. Ignored if ``nets`` is specified.
    :type n_input_units: int, required
    :param n_output_units:
        Number of output units for each neural network. Ignored if ``nets`` is specified.
    :type n_output_units: int, required
    :param batch_size:
        **[DEPRECATED and IGNORED]**
        Each batch will use all samples generated.
        Please specify ``n_batches_train`` and ``n_batches_valid`` instead.
    :type batch_size: int
    :param shuffle:
        **[DEPRECATED and IGNORED]**
        Shuffling should be performed by generators.
    :type shuffle: bool
    """

    def __init__(self, diff_eqs, conditions,
                 nets=None, train_generator=None, valid_generator=None, analytic_solutions=None,
                 optimizer=None, criterion=None, n_batches_train=1, n_batches_valid=4,
                 metrics=None, n_input_units=None, n_output_units=None,
                 # deprecated arguments are listed below
                 shuffle=None, batch_size=None):
        # deprecate argument `shuffle`
        if shuffle:
            warnings.warn(
                "param `shuffle` is deprecated and ignored; shuffling should be performed by generators",
                FutureWarning,
            )
        # deprecate argument `batch_size`
        if batch_size is not None:
            warnings.warn(
                "param `batch_size` is deprecated and ignored; specify n_batches_train and n_batches_valid instead",
                FutureWarning,
            )

        self.diff_eqs = diff_eqs
        self.conditions = conditions
        self.n_funcs = len(conditions)
        if nets is None:
            self.nets = [
                FCNN(n_input_units=n_input_units, n_output_units=n_output_units, hidden_units=(32, 32), actv=nn.Tanh)
                for _ in range(self.n_funcs)
            ]
        else:
            self.nets = nets

        if train_generator is None:
            raise ValueError("train_generator must be specified")

        if valid_generator is None:
            raise ValueError("valid_generator must be specified")

        self.metrics_fn = metrics if metrics else {}
        # For backward compatibility with the legacy `analytic_solutions` argument
        if analytic_solutions:
            warnings.warn(
                'The `analytic_solutions` argument is deprecated and could lead to unstable behavior. '
                'Pass a `metrics` dict instead.',
                FutureWarning,
            )

            def analytic_mse(*args):
                x = args[-n_input_units:]
                u_hat = analytic_solutions(*x)
                u = args[:-n_input_units]
                u, u_hat = torch.stack(u), torch.stack(u_hat)
                return ((u - u_hat) ** 2).mean()

            if 'analytic_mse' in self.metrics_fn:
                warnings.warn(
                    "Ignoring `analytic_solutions` in presence of key 'analytic_mse' in `metrics`",
                    FutureWarning,
                )
            else:
                self.metrics_fn['analytic_mse'] = analytic_mse

        # metric history, keys will be train_loss, valid_loss, train__<metric_name>, valid__<metric_name>.
        # For compatibility with ode.py and pde.py,
        # double underscore are used between 'train'/'valid' and custom metric names.
        self.metrics_history = {}
        self.metrics_history.update({'train_loss': [], 'valid_loss': []})
        self.metrics_history.update({'train__' + name: [] for name in self.metrics_fn})
        self.metrics_history.update({'valid__' + name: [] for name in self.metrics_fn})

        self.optimizer = optimizer if optimizer else Adam(chain.from_iterable(n.parameters() for n in self.nets))
        self._set_criterion(criterion)

        def make_pair_dict(train=None, valid=None):
            return {'train': train, 'valid': valid}

        self.generator = make_pair_dict(
            train=SamplerGenerator(train_generator),
            valid=SamplerGenerator(valid_generator),
        )
        # number of batches for training / validation;
        self.n_batches = make_pair_dict(train=n_batches_train, valid=n_batches_valid)
        # current batch of samples, kept for additional_loss term to use
        self._batch = make_pair_dict()
        # current network with lowest loss
        self.best_nets = None
        # current lowest loss
        self.lowest_loss = None
        # local epoch in a `.fit` call, should only be modified inside self.fit()
        self.local_epoch = 0
        # maximum local epochs to run in a `.fit()` call, should only set by inside self.fit()
        self._max_local_epoch = 0
        # controls early stopping, should be set to False at the beginning of a `.fit()` call
        # and optionally set to False by `callbacks` in `.fit()` to support early stopping
        self._stop_training = False
        # the _phase variable is registered for callback functions to access
        self._phase = None

    def _set_criterion(self, criterion):
        if criterion is None:
            self.criterion = lambda r, f, x: (r ** 2).mean()
        elif isinstance(criterion, nn.modules.loss._Loss):
            self.criterion = lambda r, f, x: criterion(r, torch.zeros_like(r))
        elif isinstance(criterion, str):
            self.criterion = _losses[criterion.lower()]
        elif callable(criterion):
            self.criterion = criterion
        else:
            raise TypeError(f"Unknown type of criterion {type(criterion)}")

    @property
    def global_epoch(self):
        r"""Global epoch count, always equal to the length of train loss history.

        :return: Number of training epochs that have been run.
        :rtype: int
        """
        return len(self.metrics_history['train_loss'])

    @property
    def batch(self):
        return self._batch

    @property
    def _batch_examples(self):
        warnings.warn(
            '`._batch_examples` has been deprecated in favor of `._batch` and will be removed in a future version',
            FutureWarning,
        )
        return self._batch

    def compute_func_val(self, net, cond, *coordinates):
        r"""Compute the function value evaluated on the points specified by ``coordinates``.

        :param net: The network to be parameterized and evaluated.
        :type net: torch.nn.Module
        :param cond: The condition (a.k.a. parameterization) for the network.
        :type cond: `neurodiffeq.conditions.BaseCondition`
        :param coordinates: A tuple of coordinate components, each with shape = (-1, 1).
        :type coordinates: tuple[torch.Tensor]
        :return: Function values at the sampled points.
        :rtype: torch.Tensor
        """
        return cond.enforce(net, *coordinates)

    def _update_history(self, value, metric_type, key):
        r"""Append a value to corresponding history list.

        :param value: Value to be appended.
        :type value: float
        :param metric_type: Name of the metric. Must be 'loss' or present in ``self.metrics``.
        :type metric_type: str
        :param key: {'train', 'valid'}. Phase of the process.
        :type key: str
        """
        self._phase = key
        if metric_type == 'loss':
            self.metrics_history[f'{key}_{metric_type}'].append(value)
        elif metric_type in self.metrics_fn:
            self.metrics_history[f'{key}__{metric_type}'].append(value)
        else:
            raise KeyError(f"metric '{metric_type}' not specified")

    def _update_train_history(self, value, metric_type):
        r"""Append a value to corresponding training history list."""
        self._update_history(value, metric_type, key='train')

    def _update_valid_history(self, value, metric_type):
        r"""Append a value to corresponding validation history list."""
        self._update_history(value, metric_type, key='valid')

    def _generate_batch(self, key):
        r"""Generate the next batch, register in self._batch and return the batch.

        :param key:
            {'train', 'valid'};
            Dict key in ``self._examples``, ``self._batch``, or ``self._batch_start``
        :type key: str
        :return: The generated batch of points.
        :type: List[`torch.Tensor`]
        """
        # the following side effects are helpful for future extension,
        # especially for additional loss term that depends on the coordinates
        self._phase = key
        self._batch[key] = [v.reshape(-1, 1) for v in self.generator[key].get_examples()]
        return self._batch[key]

    def _generate_train_batch(self):
        r"""Generate the next training batch, register in ``self._batch`` and return."""
        return self._generate_batch('train')

    def _generate_valid_batch(self):
        r"""Generate the next validation batch, register in ``self._batch`` and return."""
        return self._generate_batch('valid')

    def _do_optimizer_step(self, closure=None):
        r"""Optimization procedures after gradients have been computed. Usually ``self.optimizer.step()`` is sufficient.
        At times, users can overwrite this method to perform gradient clipping, etc. Here is an example::

            import itertools
            class MySolver(Solver)
                def _do_optimizer_step(self, closure=None):
                    nn.utils.clip_grad_norm_(itertools.chain([net.parameters() for net in self.nets]), 1.0, 'inf')
                    self.optimizer.step(closure=closure)
        """
        self.optimizer.step(closure=closure)

    def _run_epoch(self, key):
        r"""Run an epoch on train/valid points, update history, and perform an optimization step if key=='train'.

        :param key: {'train', 'valid'}; phase of the epoch
        :type key: str

        .. note::
            The optimization step is only performed after all batches are run.
        """
        if self.n_batches[key] <= 0:
            # XXX maybe we should append NaN to metric history?
            return
        self._phase = key
        epoch_loss = 0.0
        batch_loss = 0.0
        metric_values = {name: 0.0 for name in self.metrics_fn}

        # Zero the gradient only once, before running the batches. Gradients of different batches are accumulated.
        if key == 'train' and not _requires_closure(self.optimizer):
            self.optimizer.zero_grad()

        # perform forward pass for all batches: a single graph is created and release in every iteration
        # see https://discuss.pytorch.org/t/why-do-we-need-to-set-the-gradients-manually-to-zero-in-pytorch/4903/17
        for batch_id in range(self.n_batches[key]):
            batch = self._generate_batch(key)

            def closure(zero_grad=True):
                nonlocal batch_loss
                if key == 'train' and zero_grad:
                    self.optimizer.zero_grad()
                funcs = [
                    self.compute_func_val(n, c, *batch) for n, c in zip(self.nets, self.conditions)
                ]

                for name in self.metrics_fn:
                    value = self.metrics_fn[name](*funcs, *batch).item()
                    metric_values[name] += value
                residuals = self.diff_eqs(*funcs, *batch)
                residuals = torch.cat(residuals, dim=1)
                try:
                    loss = self.criterion(residuals, funcs, batch) + self.additional_loss(residuals, funcs, batch)
                except TypeError as e:
                    warnings.warn(
                        "You might need to update your code. "
                        "Since v0.4.0; both `criterion` and `additional_loss` requires three inputs: "
                        "`residual`, `funcs`, and `coords`. See documentation for more.", FutureWarning)
                    raise e

                # accumulate gradients before the current graph is collected as garbage
                if key == 'train':
                    loss.backward()
                    batch_loss = loss.item()
                return loss

            if key == 'train':
                if _requires_closure(self.optimizer):
                    # If `closure` is required by `optimizer.step()`, perform a step for every batch
                    self._do_optimizer_step(closure=closure)
                else:
                    # Otherwise, only perform backward propagation.
                    # Optimizer step will be performed only once outside the for-loop (i.e. after all batches).
                    closure(zero_grad=False)
                epoch_loss += batch_loss
            else:
                epoch_loss += closure().item()

        # calculate mean loss of all batches and register to history
        self._update_history(epoch_loss / self.n_batches[key], 'loss', key)

        # perform the optimizer step after all batches are run (if optimizer.step doesn't require `closure`)
        if key == 'train' and not _requires_closure(self.optimizer):
            self._do_optimizer_step()
        if key == 'valid':
            self._update_best()

        # calculate average metrics across batches and register to history
        for name in self.metrics_fn:
            self._update_history(
                metric_values[name] / self.n_batches[key], name, key)

    def run_train_epoch(self):
        r"""Run a training epoch, update history, and perform gradient descent."""
        self._run_epoch('train')

    def run_valid_epoch(self):
        r"""Run a validation epoch and update history."""
        self._run_epoch('valid')

    def _update_best(self):
        r"""Update ``self.lowest_loss`` and ``self.best_nets``
        if current validation loss is lower than ``self.lowest_loss``
        """
        current_loss = self.metrics_history['valid_loss'][-1]
        if (self.lowest_loss is None) or current_loss < self.lowest_loss:
            self.lowest_loss = current_loss
            self.best_nets = deepcopy(self.nets)

    def fit(self, max_epochs, callbacks=(), **kwargs):
        r"""Run multiple epochs of training and validation, update best loss at the end of each epoch.

        If ``callbacks`` is passed, callbacks are run, one at a time,
        after training and validating and updating best model but before monitor checking

        :param max_epochs: Number of epochs to run.
        :type max_epochs: int
        :param callbacks:
            A list of callback functions.
            Each function should accept the ``solver`` instance itself as its **only** argument.
        :rtype callbacks: list[callable]

        .. note::
            1. This method does not return solution, which is done in the ``.get_solution()`` method.
            2. A callback function `cb(solver)` can set ``solver._stop_training`` to True to perform early stopping.
        """
        self._stop_training = False
        self._max_local_epoch = max_epochs

        monitor = kwargs.pop('monitor', None)
        if monitor:
            warnings.warn("Passing `monitor` is deprecated, "
                          "use a MonitorCallback and pass a list of callbacks instead")
            callbacks = [monitor.to_callback()] + list(callbacks)
        if kwargs:
            raise ValueError(f'Unknown keyword argument(s): {list(kwargs.keys())}')  # pragma: no cover

        for local_epoch in range(max_epochs):
            # stop training if self._stop_training is set to True by a callback
            if self._stop_training:
                break

            # register local epoch (starting from 1 instead of 0) so it can be accessed by callbacks
            self.local_epoch = local_epoch + 1
            self.run_train_epoch()
            self.run_valid_epoch()

            for cb in callbacks:
                cb(self)

    @abstractmethod
    def get_solution(self, copy=True, best=True):
        r"""Get a (callable) solution object. See this usage example:

        .. code-block:: python3

            solution = solver.get_solution()
            point_coords = train_generator.get_examples()
            value_at_points = solution(point_coords)

        :param copy:
            Whether to make a copy of the networks so that subsequent training doesn't affect the solution;
            Defaults to True.
        :type copy: bool
        :param best:
            Whether to return the solution with lowest loss instead of the solution after the last epoch.
            Defaults to True.
        :type best: bool
        :return:
            A solution object which can be called.
            To evaluate the solution on certain points,
            you should pass the coordinates vector(s) to the returned solution.
        :rtype: BaseSolution
        """
        pass  # pragma: no cover

    def _get_internal_variables(self):
        r"""Get a dict of all available internal variables.

        :return:
            All available internal variables,
            where keys are variable names and values are the corresponding variables.
        :rtype: dict

        .. note::
            Children classes should inherit all items and optionally include new ones.
        """

        return {
            "metrics": self.metrics_fn,
            "n_batches": self.n_batches,
            "best_nets": self.best_nets,
            "criterion": self.criterion,
            "conditions": self.conditions,
            "global_epoch": self.global_epoch,
            "lowest_loss": self.lowest_loss,
            "n_funcs": self.n_funcs,
            "nets": self.nets,
            "optimizer": self.optimizer,
            "diff_eqs": self.diff_eqs,
            "generator": self.generator,
            "train_generator": self.generator['train'],
            "valid_generator": self.generator['valid'],
        }

    @deprecated_alias(param_names='var_names')
    def get_internals(self, var_names=None, return_type='list'):
        r"""Return internal variable(s) of the solver

        - If var_names == 'all', return all internal variables as a dict.
        - If var_names is single str, return the corresponding variables.
        - If var_names is a list and return_type == 'list', return corresponding internal variables as a list.
        - If var_names is a list and return_type == 'dict', return a dict with keys in var_names.

        :param var_names: An internal variable name or a list of internal variable names.
        :type var_names: str or list[str]
        :param return_type: {'list', 'dict'}; Ignored if ``var_names`` is a string.
        :type return_type: str
        :return: A single variable, or a list/dict of internal variables as indicated above.
        :rtype: list or dict or any
        """

        available_variables = self._get_internal_variables()

        if var_names == "all" or var_names is None:
            return available_variables

        if isinstance(var_names, str):
            return available_variables[var_names]

        if return_type == 'list':
            return [available_variables[name] for name in var_names]
        elif return_type == "dict":
            return {name: available_variables[name] for name in var_names}
        else:
            raise ValueError(f"unrecognized return_type = {return_type}")

    def additional_loss(self, residual, funcs, coords):
        r"""Additional loss terms for training. This method is to be overridden by subclasses.
        This method can use any of the internal variables: self.nets, self.conditions, self.global_epoch, etc.

        :param residual: Residual tensor of differential equation. It has shape (N_SAMPLES, N_EQUATIONS)
        :type residual: torch.Tensor
        :param funcs:
            Outputs of the networks after parameterization.
            There are ``len(nets)`` entries in total. Each entry is a tensor of shape (N_SAMPLES, N_OUTPUT_UNITS).
        :type funcs: List[torch.Tensor]
        :param coords:
            Inputs to the networks; a.k.a. the spatio-temporal coordinates of the system.
            There are ``N_COORDS`` entries in total. Each entry is a tensor of shape (N_SAMPLES, 1).
        :type coords: List[torch.Tensor]
        :return: Additional loss. Must be a ``torch.Tensor`` of empty shape (scalar).
        :rtype: torch.Tensor
        """
        return 0.0


class BaseSolution(ABC):
    r"""A solution to a PDE/ODE (system).

    :param nets:
        The neural networks that approximate the PDE/ODE solution.

        - If ``nets`` is a list of ``torch.nn.Module``, it should have the same length with ``conditions``
        - If ``nets`` is a single ``torch.nn.Module``, it should have as many output units as length of ``conditions``

    :type nets: list[`torch.nn.Module`] or `torch.nn.Module`
    :param conditions:
        A list of conditions that should be enforced on the PDE/ODE solution.
        ``conditions`` should have a length equal to the number of dependent variables in the ODE/PDE system.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    """

    def __init__(self, nets, conditions):
        if isinstance(nets, nn.Module):
            # This is for backward compatibility with the `single_net` option
            # The same torch.nn.Module instance is repeated to form a list of the same length as `conditions`
            self.nets = [nets] * len(conditions)
        else:
            self.nets = nets
        self.conditions = conditions

    @abstractmethod
    def _compute_u(self, net, condition, *coords):
        pass  # pragma: no cover

    @deprecated_alias(as_type='to_numpy')
    def __call__(self, *coords, to_numpy=False):
        r"""Evaluate the solution at certain points.

        :param coords: tuple of coordinate tensors, each of shape (n_samples, 1)
        :type coords: Tuple[`torch.Tensor`]
        :param to_numpy:
            If set to True, the call returns a ``numpy.ndarray`` instead of ``torch.Tensor``.
            Defaults to False.
        :type to_numpy: bool
        :return: Dependent variables evaluated at given points.
        :rtype: list[`torch.Tensor` or `numpy.array`] or `torch.Tensor` or `numpy.array`
        """
        coords = [c if isinstance(c, torch.Tensor) else torch.tensor(c) for c in coords]
        original_shape = coords[0].shape
        coords = [c.reshape(-1, 1) for c in coords]
        if isinstance(to_numpy, str):
            # Why did we allow `tf` as an option >_<
            # We should phase this out as soon as possible
            if to_numpy == 'tf' or to_numpy == 'torch':
                to_numpy = False
            elif to_numpy == 'np':
                to_numpy = True
            else:
                raise ValueError(f"Unrecognized `as_type` option: '{to_numpy}'")

        us = [
            self._compute_u(net, con, *coords).reshape(original_shape)
            for con, net in zip(self.conditions, self.nets)
        ]
        if to_numpy:
            us = [u.detach().cpu().numpy() for u in us]

        return us if len(self.nets) > 1 else us[0]


class GenericSolution(BaseSolution):
    def _compute_u(self, net, condition, *coords):
        return condition.enforce(net, *coords)


class GenericSolver(BaseSolver):
    def get_solution(self, copy=True, best=True):
        r"""Get a (callable) solution object. See this usage example:

        .. code-block:: python3

            solution = solver.get_solution()
            point_coords = train_generator.get_examples()
            value_at_points = solution(point_coords)

        :param copy:
            Whether to make a copy of the networks so that subsequent training doesn't affect the solution;
            Defaults to True.
        :type copy: bool
        :param best:
            Whether to return the solution with lowest loss instead of the solution after the last epoch.
            Defaults to True.
        :type best: bool
        :return:
            A solution object which can be called.
            To evaluate the solution on certain points,
            you should pass the coordinates vector(s) to the returned solution.
        :rtype: BaseSolution
        """
        nets = self.best_nets if best else self.nets
        conditions = self.conditions
        if copy:
            nets = deepcopy(nets)
            conditions = deepcopy(conditions)

        return GenericSolution(nets, conditions)


class SolverSpherical(BaseSolver):
    r"""A solver class for solving PDEs in spherical coordinates

    :param pde_system:
        The PDE system to solve, which maps a tuple of three coordinates to a tuple of PDE residuals,
        both the coordinates and PDE residuals must have shape (n_samples, 1).
    :type pde_system: callable
    :param conditions:
        List of boundary conditions for each target function.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    :param r_min:
        Radius for inner boundary (:math:`r_0>0`).
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type r_min: float, optional
    :param r_max:
        Radius for outer boundary (:math:`r_1>r_0`).
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type r_max: float, optional
    :param nets:
        List of neural networks for parameterized solution.
        If provided, length of ``nets`` must equal that of ``conditions``
    :type nets: list[torch.nn.Module], optional
    :param train_generator:
        Generator for sampling training points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``train_generator`` must be specified if ``r_min`` and ``r_max`` are not set.
    :type train_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param valid_generator:
        Generator for sampling validation points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``valid_generator`` must be specified if ``r_min`` and ``r_max`` are not set.
    :type valid_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param analytic_solutions:
        Analytical solutions to be compared with neural net solutions.
        It maps a tuple of three coordinates to a tuple of function values.
        Output shape should match that of ``nets``.
    :type analytic_solutions: callable, optional
    :param optimizer:
        Optimizer to be used for training.
        Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
    :type optimizer: ``torch.nn.optim.Optimizer``, optional
    :param criterion:
        The loss function used for training.

        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
        - If any other callable, it must map
          A) a residual tensor (shape `(n_points, n_equations)`),
          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
          so that backpropagation can be performed.

    :type criterion:
        str or `torch.nn.moduesl.loss._Loss` or callable
    :param n_batches_train:
        Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
        Defaults to 1.
    :type n_batches_train: int, optional
    :param n_batches_valid:
        Number of batches to validate in every epoch, where batch-size equals ``valid_generator.size``.
        Defaults to 4.
    :type n_batches_valid: int, optional
    :param metrics:
        Additional metrics to be logged (besides loss). ``metrics`` should be a dict where

        - Keys are metric names (e.g. 'analytic_mse');
        - Values are functions (callables) that computes the metric value.
          These functions must accept the same input as the differential equation ``diff_eq``.

    :type metrics: dict, optional
    :param enforcer:
        A function of signature
        ``enforcer(net: nn.Module, cond: neurodiffeq.conditions.BaseCondition,
        coords: Tuple[torch.Tensor]) -> torch.Tensor``
        that returns the dependent variable value evaluated on the batch.
    :type enforcer: callable
    :param n_output_units:
        Number of output units for each neural network.
        Ignored if ``nets`` is specified.
        Defaults to 1.
    :type n_output_units: int, optional
    :param batch_size:
        **[DEPRECATED and IGNORED]**
        Each batch will use all samples generated.
        Please specify ``n_batches_train`` and ``n_batches_valid`` instead.
    :type batch_size: int
    :param shuffle:
        **[DEPRECATED and IGNORED]**
        Shuffling should be performed by generators.
    :type shuffle: bool
    """

    def __init__(self, pde_system, conditions, r_min=None, r_max=None,
                 nets=None, train_generator=None, valid_generator=None, analytic_solutions=None,
                 optimizer=None, criterion=None, n_batches_train=1, n_batches_valid=4, metrics=None, enforcer=None,
                 n_output_units=1,
                 # deprecated arguments are listed below
                 shuffle=None, batch_size=None):

        if train_generator is None or valid_generator is None:
            if r_min is None or r_max is None:
                raise ValueError(f"Either generator is not provided, r_min and r_max should be both provided: "
                                 f"got r_min={r_min}, r_max={r_max}, train_generator={train_generator}, "
                                 f"valid_generator={valid_generator}")

        if train_generator is None:
            train_generator = GeneratorSpherical(512, r_min, r_max, method='equally-spaced-noisy')

        if valid_generator is None:
            valid_generator = GeneratorSpherical(512, r_min, r_max, method='equally-spaced-noisy')

        self.r_min, self.r_max = r_min, r_max
        self.enforcer = enforcer

        super(SolverSpherical, self).__init__(
            diff_eqs=pde_system,
            conditions=conditions,
            nets=nets,
            train_generator=train_generator,
            valid_generator=valid_generator,
            analytic_solutions=analytic_solutions,
            optimizer=optimizer,
            criterion=criterion,
            n_batches_train=n_batches_train,
            n_batches_valid=n_batches_valid,
            metrics=metrics,
            n_input_units=3,
            n_output_units=n_output_units,
            shuffle=shuffle,
            batch_size=batch_size,
        )

    def _auto_enforce(self, net, cond, *coordinates):
        r"""Enforce condition on network with inputs. If self.enforcer is set, use it.
        Otherwise, fill cond.enforce() with as many arguments as needed.

        :param net: Network for parameterized solution.
        :type net: torch.nn.Module
        :param cond: Condition (a.k.a. parameterization) for the network.
        :type cond: `neurodiffeq.conditions.BaseCondition`
        :param coordinates: A tuple of vectors, each with shape = (-1, 1).
        :type coordinates: tuple[torch.Tensor]
        :return: Function values at sampled points.
        :rtype: torch.Tensor
        """
        if self.enforcer:
            return self.enforcer(net, cond, coordinates)

        # Base .enforce takes a variable length *arg; n_params should be deduced from .parameterize
        if cond.__class__.enforce == BaseCondition.enforce:
            n_params = len(signature(cond.parameterize).parameters)
        else:
            n_params = len(signature(cond.enforce).parameters)
        coordinates = coordinates[:n_params - 1]
        return cond.enforce(net, *coordinates)

    def compute_func_val(self, net, cond, *coordinates):
        r"""Enforce condition on network with inputs. If self.enforcer is set, use it.
        Otherwise, fill cond.enforce() with as many arguments as needed.

        :param net: Network for parameterized solution.
        :type net: torch.nn.Module
        :param cond: Condition (a.k.a. parameterization) for the network.
        :type cond: `neurodiffeq.conditions.BaseCondition`
        :param coordinates: A tuple of vectors, each with shape = (-1, 1).
        :type coordinates: tuple[torch.Tensor]
        :return: Function values at sampled points.
        :rtype: torch.Tensor
        """
        return self._auto_enforce(net, cond, *coordinates)

    def get_solution(self, copy=True, best=True, harmonics_fn=None):
        r"""Get a (callable) solution object. See this usage example:

        .. code-block:: python3

            solution = solver.get_solution()
            point_coords = train_generator.get_examples()
            value_at_points = solution(point_coords)

        :param copy:
            Whether to make a copy of the networks so that subsequent training doesn't affect the solution;
            Defaults to True.
        :type copy: bool
        :param best:
            Whether to return the solution with lowest loss instead of the solution after the last epoch.
            Defaults to True.
        :type best: bool
        :param harmonics_fn:
            If set, use it as function basis for returned solution.
        :type harmonics_fn: callable
        :return: The solution after training.
        :rtype: ``neurodiffeq.solvers.BaseSolution``
        """
        nets = self.best_nets if best else self.nets
        conditions = self.conditions
        if copy:
            nets = deepcopy(nets)
            conditions = deepcopy(conditions)

        if harmonics_fn:
            return SolutionSphericalHarmonics(nets, conditions, harmonics_fn=harmonics_fn)
        else:
            return SolutionSpherical(nets, conditions)

    def _get_internal_variables(self):
        available_variables = super(SolverSpherical, self)._get_internal_variables()
        available_variables.update({
            'r_min': self.r_min,
            'r_max': self.r_max,
            'enforcer': self.enforcer,
        })
        return available_variables


class SolutionSpherical(BaseSolution):
    def _compute_u(self, net, condition, rs, thetas, phis):
        return condition.enforce(net, rs, thetas, phis)


class SolutionSphericalHarmonics(SolutionSpherical):
    r"""A solution to a PDE (system) in spherical coordinates.

    :param nets: List of networks that takes in radius tensor and outputs the coefficients of spherical harmonics.
    :type nets: list[`torch.nn.Module`]
    :param conditions: List of conditions to be enforced on each nets; must be of the same length as nets.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    :param harmonics_fn: Mapping from :math:`\theta` and :math:`\phi` to basis functions, e.g., spherical harmonics.
    :type harmonics_fn: callable
    :param max_degree: **DEPRECATED and SUPERSEDED** by ``harmonics_fn``. Highest used for the harmonic basis.
    :type max_degree: int
    """

    def __init__(self, nets, conditions, max_degree=None, harmonics_fn=None):
        super(SolutionSphericalHarmonics, self).__init__(nets, conditions)
        if (harmonics_fn is None) and (max_degree is None):
            raise ValueError("harmonics_fn should be specified")

        if max_degree is not None:
            warnings.warn(
                "`max_degree` is DEPRECATED; pass `harmonics_fn` instead, which takes precedence",
                FutureWarning,
            )
            self.harmonics_fn = RealSphericalHarmonics(max_degree=max_degree)

        if harmonics_fn is not None:
            self.harmonics_fn = harmonics_fn

    def _compute_u(self, net, condition, rs, thetas, phis):
        products = condition.enforce(net, rs) * self.harmonics_fn(thetas, phis)
        return torch.sum(products, dim=1)


class Solution1D(BaseSolution):
    def _compute_u(self, net, condition, ts):
        return condition.enforce(net, ts)


class Solver1D(BaseSolver):
    r"""A solver class for solving ODEs (single-input differential equations)

    :param ode_system:
        The ODE system to solve, which maps a torch.Tensor to a tuple of ODE residuals,
        both the input and output must have shape (n_samples, 1).
    :type ode_system: callable
    :param conditions:
        List of conditions for each target function.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    :param t_min:
        Lower bound of input (start time).
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type t_min: float, optional
    :param t_max:
        Upper bound of input (start time).
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type t_max: float, optional
    :param nets:
        List of neural networks for parameterized solution.
        If provided, length of ``nets`` must equal that of ``conditions``
    :type nets: list[torch.nn.Module], optional
    :param train_generator:
        Generator for sampling training points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``train_generator`` must be specified if ``t_min`` and ``t_max`` are not set.
    :type train_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param valid_generator:
        Generator for sampling validation points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``valid_generator`` must be specified if ``t_min`` and ``t_max`` are not set.
    :type valid_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param analytic_solutions:
        Analytical solutions to be compared with neural net solutions.
        It maps a torch.Tensor to a tuple of function values.
        Output shape should match that of ``nets``.
    :type analytic_solutions: callable, optional
    :param optimizer:
        Optimizer to be used for training.
        Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
    :type optimizer: ``torch.nn.optim.Optimizer``, optional
    :param criterion:
        The loss function used for training.

        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
        - If any other callable, it must map
          A) a residual tensor (shape `(n_points, n_equations)`),
          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
          so that backpropagation can be performed.

    :type criterion:
        str or `torch.nn.moduesl.loss._Loss` or callable
    :param n_batches_train:
        Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
        Defaults to 1.
    :type n_batches_train: int, optional
    :param n_batches_valid:
        Number of batches to validate in every epoch, where batch-size equals ``valid_generator.size``.
        Defaults to 4.
    :type n_batches_valid: int, optional
    :param metrics:
        Additional metrics to be logged (besides loss). ``metrics`` should be a dict where

        - Keys are metric names (e.g. 'analytic_mse');
        - Values are functions (callables) that computes the metric value.
          These functions must accept the same input as the differential equation ``ode_system``.

    :type metrics: dict[str, callable], optional
    :param n_output_units:
        Number of output units for each neural network.
        Ignored if ``nets`` is specified.
        Defaults to 1.
    :type n_output_units: int, optional
    :param batch_size:
        **[DEPRECATED and IGNORED]**
        Each batch will use all samples generated.
        Please specify ``n_batches_train`` and ``n_batches_valid`` instead.
    :type batch_size: int
    :param shuffle:
        **[DEPRECATED and IGNORED]**
        Shuffling should be performed by generators.
    :type shuffle: bool
    """

    def __init__(self, ode_system, conditions, t_min=None, t_max=None,
                 nets=None, train_generator=None, valid_generator=None, analytic_solutions=None, optimizer=None,
                 criterion=None, n_batches_train=1, n_batches_valid=4, metrics=None, n_output_units=1,
                 # deprecated arguments are listed below
                 batch_size=None, shuffle=None):

        if train_generator is None or valid_generator is None:
            if t_min is None or t_max is None:
                raise ValueError(f"Either generator is not provided, t_min and t_max should be both provided: \n"
                                 f"got t_min={t_min}, t_max={t_max}, "
                                 f"train_generator={train_generator}, valid_generator={valid_generator}")

        if train_generator is None:
            train_generator = Generator1D(32, t_min=t_min, t_max=t_max, method='equally-spaced-noisy')
        if valid_generator is None:
            valid_generator = Generator1D(32, t_min=t_min, t_max=t_max, method='equally-spaced')

        self.t_min, self.t_max = t_min, t_max

        super(Solver1D, self).__init__(
            diff_eqs=ode_system,
            conditions=conditions,
            nets=nets,
            train_generator=train_generator,
            valid_generator=valid_generator,
            analytic_solutions=analytic_solutions,
            optimizer=optimizer,
            criterion=criterion,
            n_batches_train=n_batches_train,
            n_batches_valid=n_batches_valid,
            metrics=metrics,
            n_input_units=1,
            n_output_units=n_output_units,
            shuffle=shuffle,
            batch_size=batch_size,
        )

    def get_solution(self, copy=True, best=True):
        r"""Get a (callable) solution object. See this usage example:

        .. code-block:: python3

            solution = solver.get_solution()
            point_coords = train_generator.get_examples()
            value_at_points = solution(point_coords)

        :param copy:
            Whether to make a copy of the networks so that subsequent training doesn't affect the solution;
            Defaults to True.
        :type copy: bool
        :param best:
            Whether to return the solution with lowest loss instead of the solution after the last epoch.
            Defaults to True.
        :type best: bool
        :return:
            A solution object which can be called.
            To evaluate the solution on certain points,
            you should pass the coordinates vector(s) to the returned solution.
        :rtype: BaseSolution
        """
        nets = self.best_nets if best else self.nets
        conditions = self.conditions
        if copy:
            nets = deepcopy(nets)
            conditions = deepcopy(conditions)

        return Solution1D(nets, conditions)

    def _get_internal_variables(self):
        available_variables = super(Solver1D, self)._get_internal_variables()
        available_variables.update({
            't_min': self.t_min,
            't_max': self.t_max,
        })
        return available_variables


class BundleSolution1D(BaseSolution):
    def _compute_u(self, net, condition, *ts):
        return condition.enforce(net, *ts)


class BundleSolver1D(BaseSolver):
    r"""A solver class for solving ODEs (single-input differential equations)
    , or a bundle of ODEs for different values of its parameters and/or conditions

    :param ode_system:
        The ODE system to solve, which maps a torch.Tensor or a tuple of torch.Tensors, to a tuple of ODE residuals,
        both the input and output must have shape (n_samples, 1).
    :type ode_system: callable
    :param conditions:
        List of conditions for each target function.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    :param t_min:
        Lower bound of input (start time).
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type t_min: float, optional
    :param t_max:
        Upper bound of input (start time).
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type t_max: float, optional
    :param theta_min:
        Lower bound of input (parameters and/or conditions). If conditions are included in the bundle,
        the order should match the one inferred by the values of the ``bundle_conditions`` input
        in the ``neurodiffeq.conditions.BundleIVP``.
        Defaults to None.
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type theta_min: float or tuple, optional
    :param theta_max:
        Upper bound of input (parameters and/or conditions). If conditions are included in the bundle,
        the order should match the one inferred by the values of the ``bundle_conditions`` input
        in the ``neurodiffeq.conditions.BundleIVP``.
        Defaults to None.
        Ignored if ``train_generator`` and ``valid_generator`` are both set.
    :type theta_max: float or tuple, optional
    :param nets:
        List of neural networks for parameterized solution.
        If provided, length of ``nets`` must equal that of ``conditions``
    :type nets: list[torch.nn.Module], optional
    :param train_generator:
        Generator for sampling training points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``train_generator`` must be specified if ``t_min`` and ``t_max`` are not set.
    :type train_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param valid_generator:
        Generator for sampling validation points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``valid_generator`` must be specified if ``t_min`` and ``t_max`` are not set.
    :type valid_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param analytic_solutions:
        Analytical solutions to be compared with neural net solutions.
        It maps a torch.Tensor to a tuple of function values.
        Output shape should match that of ``nets``.
    :type analytic_solutions: callable, optional
    :param optimizer:
        Optimizer to be used for training.
        Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
    :type optimizer: ``torch.nn.optim.Optimizer``, optional
    :param criterion:
        The loss function used for training.

        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
        - If any other callable, it must map
          A) a residual tensor (shape `(n_points, n_equations)`),
          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
          so that backpropagation can be performed.

    :type criterion:
        str or `torch.nn.moduesl.loss._Loss` or callable
    :param n_batches_train:
        Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
        Defaults to 1.
    :type n_batches_train: int, optional
    :param n_batches_valid:
        Number of batches to validate in every epoch, where batch-size equals ``valid_generator.size``.
        Defaults to 4.
    :type n_batches_valid: int, optional
    :param metrics:
        Additional metrics to be logged (besides loss). ``metrics`` should be a dict where

        - Keys are metric names (e.g. 'analytic_mse');
        - Values are functions (callables) that computes the metric value.
          These functions must accept the same input as the differential equation ``ode_system``.

    :type metrics: dict[str, callable], optional
    :param n_output_units:
        Number of output units for each neural network.
        Ignored if ``nets`` is specified.
        Defaults to 1.
    :type n_output_units: int, optional
    :param batch_size:
        **[DEPRECATED and IGNORED]**
        Each batch will use all samples generated.
        Please specify ``n_batches_train`` and ``n_batches_valid`` instead.
    :type batch_size: int
    :param shuffle:
        **[DEPRECATED and IGNORED]**
        Shuffling should be performed by generators.
    :type shuffle: bool
    """

    def __init__(self, ode_system, conditions, t_min, t_max,
                 theta_min=None, theta_max=None,
                 nets=None, train_generator=None, valid_generator=None, analytic_solutions=None, optimizer=None,
                 criterion=None, n_batches_train=1, n_batches_valid=4, metrics=None, n_output_units=1,
                 # deprecated arguments are listed below
                 batch_size=None, shuffle=None):

        if train_generator is None or valid_generator is None:
            if t_min is None or t_max is None:
                raise ValueError(f"Either generator is not provided, t_min and t_max should be both provided: \n"
                                 f"got t_min={t_min}, t_max={t_max}, "
                                 f"train_generator={train_generator}, valid_generator={valid_generator}")

        if isinstance(theta_min, (float, int)):
            theta_min = (theta_min,)

        if isinstance(theta_max, (float, int)):
            theta_max = (theta_max,)

        if theta_min is None and theta_max is None:
            r_min = (t_min,)
            r_max = (t_max,)
        else:
            r_min = (t_min,) + theta_min
            r_max = (t_max,) + theta_max

        n_input_units = len(r_min)

        if train_generator is None:
            train_generator = Generator1D(32, t_min=t_min, t_max=t_max, method='equally-spaced-noisy')
            for i in range(n_input_units - 1):
                train_generator ^= Generator1D(32, t_min=r_min[i + 1], t_max=r_max[i + 1],
                                               method='equally-spaced-noisy')
        if valid_generator is None:
            valid_generator = Generator1D(32, t_min=t_min, t_max=t_max, method='equally-spaced')
            for i in range(n_input_units - 1):
                valid_generator ^= Generator1D(32, t_min=r_min[i + 1], t_max=r_max[i + 1], method='equally-spaced')

        non_var = []
        for c in conditions:
            try:
                bc = tuple(c.bundle_conditions.values())
            except Exception:
                bc = ()
            for index in bc:
                non_var.append(index)
        non_var = list(set(non_var))

        def non_var_filter(*variables):
            variables = list(variables)
            for i, n in enumerate(non_var):
                del variables[len(conditions) + 1 + n - i]
            return ode_system(*variables)

        self.r_min, self.r_max = r_min, r_max

        super(BundleSolver1D, self).__init__(
            diff_eqs=non_var_filter,
            conditions=conditions,
            nets=nets,
            train_generator=train_generator,
            valid_generator=valid_generator,
            analytic_solutions=analytic_solutions,
            optimizer=optimizer,
            criterion=criterion,
            n_batches_train=n_batches_train,
            n_batches_valid=n_batches_valid,
            metrics=metrics,
            n_input_units=n_input_units,
            n_output_units=n_output_units,
            shuffle=shuffle,
            batch_size=batch_size,
        )

    def get_solution(self, copy=True, best=True):
        r"""Get a (callable) solution object. See this usage example:

        .. code-block:: python3

            solution = solver.get_solution()
            point_coords = train_generator.get_examples()
            value_at_points = solution(point_coords)

        :param copy:
            Whether to make a copy of the networks so that subsequent training doesn't affect the solution;
            Defaults to True.
        :type copy: bool
        :param best:
            Whether to return the solution with lowest loss instead of the solution after the last epoch.
            Defaults to True.
        :type best: bool
        :return:
            A solution object which can be called.
            To evaluate the solution on certain points,
            you should pass the coordinates vector(s) to the returned solution.
        :rtype: BaseSolution
        """
        nets = self.best_nets if best else self.nets
        conditions = self.conditions
        if copy:
            nets = deepcopy(nets)
            conditions = deepcopy(conditions)

        return BundleSolution1D(nets, conditions)

    def _get_internal_variables(self):
        available_variables = super(BundleSolver1D, self)._get_internal_variables()
        available_variables.update({
            'r_min': self.r_min,
            'r_max': self.r_max,
        })
        return available_variables


class Solution2D(BaseSolution):
    def _compute_u(self, net, condition, xs, ys):
        return condition.enforce(net, xs, ys)


class Solver2D(BaseSolver):
    r"""A solver class for solving PDEs in 2 dimensions.

    :param pde_system:
        The PDE system to solve, which maps two ``torch.Tensor``s to PDE residuals (``tuple[torch.Tensor]``),
        both the input and output must have shape (n_samples, 1).
    :type pde_system: callable
    :param conditions:
        List of conditions for each target function.
    :type conditions: list[`neurodiffeq.conditions.BaseCondition`]
    :param xy_min:
        The lower bound of 2 dimensions.
        If we only care about :math:`x \geq x_0` and :math:`y \geq y_0`,
        then `xy_min` is `(x_0, y_0)`.
        Only needed when train_generator or valid_generator are not specified.
        Defaults to None
    :type xy_min: tuple[float, float], optional
    :param xy_max:
        The upper bound of 2 dimensions.
        If we only care about :math:`x \leq x_1` and :math:`y \leq y_1`, then `xy_min` is `(x_1, y_1)`.
        Only needed when train_generator or valid_generator are not specified.
        Defaults to None
    :type xy_max: tuple[float, float], optional
    :param nets:
        List of neural networks for parameterized solution.
        If provided, length of ``nets`` must equal that of ``conditions``
    :type nets: list[torch.nn.Module], optional
    :param train_generator:
        Generator for sampling training points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``train_generator`` must be specified if ``t_min`` and ``t_max`` are not set.
    :type train_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param valid_generator:
        Generator for sampling validation points,
        which must provide a ``.get_examples()`` method and a ``.size`` field.
        ``valid_generator`` must be specified if ``t_min`` and ``t_max`` are not set.
    :type valid_generator: `neurodiffeq.generators.BaseGenerator`, optional
    :param analytic_solutions:
        Analytical solutions to be compared with neural net solutions.
        It maps a torch.Tensor to a tuple of function values.
        Output shape should match that of ``nets``.
    :type analytic_solutions: callable, optional
    :param optimizer:
        Optimizer to be used for training.
        Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
    :type optimizer: ``torch.nn.optim.Optimizer``, optional
    :param criterion:
        The loss function used for training.

        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
        - If any other callable, it must map
          A) a residual tensor (shape `(n_points, n_equations)`),
          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
          so that backpropagation can be performed.

    :type criterion:
        str or `torch.nn.moduesl.loss._Loss` or callable
    :param n_batches_train:
        Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
        Defaults to 1.
    :type n_batches_train: int, optional
    :param n_batches_valid:
        Number of batches to validate in every epoch, where batch-size equals ``valid_generator.size``.
        Defaults to 4.
    :type n_batches_valid: int, optional
    :param metrics:
        Additional metrics to be logged (besides loss). ``metrics`` should be a dict where

        - Keys are metric names (e.g. 'analytic_mse');
        - Values are functions (callables) that computes the metric value.
          These functions must accept the same input as the differential equation ``ode_system``.

    :type metrics: dict[str, callable], optional
    :param n_output_units:
        Number of output units for each neural network.
        Ignored if ``nets`` is specified.
        Defaults to 1.
    :type n_output_units: int, optional
    :param batch_size:
        **[DEPRECATED and IGNORED]**
        Each batch will use all samples generated.
        Please specify ``n_batches_train`` and ``n_batches_valid`` instead.
    :type batch_size: int
    :param shuffle:
        **[DEPRECATED and IGNORED]**
        Shuffling should be performed by generators.
    :type shuffle: bool
    """

    def __init__(self, pde_system, conditions, xy_min=None, xy_max=None,
                 nets=None, train_generator=None, valid_generator=None, analytic_solutions=None, optimizer=None,
                 criterion=None, n_batches_train=1, n_batches_valid=4, metrics=None, n_output_units=1,
                 # deprecated arguments are listed below
                 batch_size=None, shuffle=None):

        if train_generator is None or valid_generator is None:
            if xy_min is None or xy_max is None:
                raise ValueError(f"Either generator is not provided, xy_min and xy_max should be both provided: \n"
                                 f"got xy_min={xy_min}, xy_max={xy_max}, "
                                 f"train_generator={train_generator}, valid_generator={valid_generator}")

        if train_generator is None:
            train_generator = Generator2D((32, 32), xy_min=xy_min, xy_max=xy_max, method='equally-spaced-noisy')
        if valid_generator is None:
            valid_generator = Generator2D((32, 32), xy_min=xy_min, xy_max=xy_max, method='equally-spaced')

        self.xy_min, self.xy_max = xy_min, xy_max

        super(Solver2D, self).__init__(
            diff_eqs=pde_system,
            conditions=conditions,
            nets=nets,
            train_generator=train_generator,
            valid_generator=valid_generator,
            analytic_solutions=analytic_solutions,
            optimizer=optimizer,
            criterion=criterion,
            n_batches_train=n_batches_train,
            n_batches_valid=n_batches_valid,
            metrics=metrics,
            n_input_units=2,
            n_output_units=n_output_units,
            shuffle=shuffle,
            batch_size=batch_size,
        )

    def get_solution(self, copy=True, best=True):
        r"""Get a (callable) solution object. See this usage example:

        .. code-block:: python3

            solution = solver.get_solution()
            point_coords = train_generator.get_examples()
            value_at_points = solution(point_coords)

        :param copy:
            Whether to make a copy of the networks so that subsequent training doesn't affect the solution;
            Defaults to True.
        :type copy: bool
        :param best:
            Whether to return the solution with lowest loss instead of the solution after the last epoch.
            Defaults to True.
        :type best: bool
        :return:
            A solution object which can be called.
            To evaluate the solution on certain points,
            you should pass the coordinates vector(s) to the returned solution.
        :rtype: BaseSolution
        """
        nets = self.best_nets if best else self.nets
        conditions = self.conditions
        if copy:
            nets = deepcopy(nets)
            conditions = deepcopy(conditions)

        return Solution2D(nets, conditions)

    def _get_internal_variables(self):
        available_variables = super(Solver2D, self)._get_internal_variables()
        available_variables.update({
            'xy_min': self.xy_min,
            'xy_max': self.xy_max,
        })
        return available_variables



import torch
from ..networks import FCNN
from ..utils import vstack, hstack, split_columns


class DiscreteSolution1D:
    def __init__(self, ts, *us):
        self.ts = ts
        self.us_tuple = torch.stack(us, dim=1)

    def __call__(self, ts):
        ret_u = []
        for t in ts:
            for i in range(len(self.ts) - 1):
                if (self.ts[i] <= t) and (t <= self.ts[i + 1]):
                    u = (self.us_tuple[i + 1] * (t - self.ts[i]) + self.us_tuple[i] * (self.ts[i + 1] - t)) \
                        / (self.ts[i + 1] - self.ts[i])
                    ret_u.append(u)
                    break

        ret_u = torch.stack(ret_u, dim=0)
        return [ret_u[:, j] for j in range(ret_u.shape[1])]


class Hypersolver:
    def __init__(self, func, u0, t0, tn, n_steps, sol, numerical_solver, net=None, optimizer=None):
        self.func = func
        if isinstance(u0, (int, float)):
            self.u0 = torch.tensor([float(u0)])
        elif isinstance(u0, (list, tuple)):
            self.u0 = torch.tensor(u0)
        else:
            raise TypeError(f"u0 must be int, float, list, or tuple, not {type(u0)}")
        self.t0 = t0
        self.tn = tn
        self.n_steps = n_steps
        self.h = (tn - t0) / n_steps
        self.ts = torch.linspace(t0, tn, n_steps + 1)
        self.solution = sol
        self.numerical_solver = numerical_solver
        self.us = torch.stack(self.solution(self.ts), dim=1)
        self.local_epoch = 0
        self._max_local_epoch = 1

        us_no_head = self.us[1:, :]
        us_no_tail = self.us[:-1, :]
        R = us_no_head - us_no_tail - \
            self.h * hstack(self.numerical_solver.step(self.func, split_columns(us_no_tail), self.ts[:-1], self.h))
        self.residual = R / self.h ** (self.numerical_solver.order + 1)

        if net is None:
            self.net = FCNN(n_input_units=len(self.u0) + 1, n_output_units=len(self.u0), hidden_units=(32, 32))
        else:
            self.net = net
        if optimizer is None:
            self.optimizer = torch.optim.Adam(self.net.parameters(), lr=1e-3)
        else:
            self.optimizer = optimizer
        self.loss_fcn = torch.nn.MSELoss()

        self.metrics_history = {}
        self.metrics_history['train_loss'] = []
        self.metrics_history['valid_loss'] = []

    def fit(self, max_epochs):
        self._max_local_epoch = max_epochs
        for epoch in range(max_epochs):
            self.local_epoch += 1
            input = torch.cat((self.ts.reshape(-1, 1), self.us), dim=1)
            output = self.net(input)
            loss = self.loss_fcn(self.residual, output[1:])

            loss.backward()
            self.optimizer.step()
            self.optimizer.zero_grad()
            self.metrics_history['train_loss'].append(loss.item())

    @property
    def global_epoch(self):
        return len(self.metrics_history['train_loss'])

    def get_solution(self):
        ret = self.numerical_solver.solve(self.func, self.u0, self.t0, self.tn, self.n_steps, hypernet=self.net)
        return DiscreteSolution1D(*ret)



import numpy as np
import pytest
from neurodiffeq.conditions import NoCondition
from neurodiffeq.networks import FCNN
from neurodiffeq.monitors import Monitor2D, MetricsMonitor, Monitor1D, StreamPlotMonitor2D

N_FUNCTIONS = 2
EMPTY_HISTORY = dict(train_loss=[], valid_loss=[])
HISTORY_10 = dict(
    train_loss=list(np.random.random(10)),
    valid_loss=list(np.random.random(10)),
)


@pytest.mark.parametrize(argnames='history', argvalues=[EMPTY_HISTORY, HISTORY_10])
@pytest.mark.parametrize(argnames='solution_style', argvalues=['curves', 'heatmap'])
def test_monitor_2d(solution_style, history):
    monitor = Monitor2D((0, 0), (1, 1), check_every=100, solution_style=solution_style)
    nets = [FCNN(2, 1) for _ in range(N_FUNCTIONS)]
    conditions = [NoCondition() for _ in range(N_FUNCTIONS)]
    monitor.check(nets, conditions, history)


@pytest.mark.parametrize(argnames='history', argvalues=[EMPTY_HISTORY, HISTORY_10])
def test_metrics_mointor(history):
    monitor = MetricsMonitor(check_every=10)
    nets = [FCNN(2, 1) for _ in range(N_FUNCTIONS)]
    conditions = [NoCondition() for _ in range(N_FUNCTIONS)]
    monitor.check(nets, conditions, history)


@pytest.mark.parametrize(argnames='history', argvalues=[EMPTY_HISTORY, HISTORY_10])
def test_monitor_1d(history):
    monitor = Monitor1D(0, 1)
    nets = [FCNN() for _ in range(N_FUNCTIONS)]
    conditions = [NoCondition() for _ in range(N_FUNCTIONS)]
    monitor.check(nets, conditions, history=history)


@pytest.mark.parametrize(argnames='specify_field_names', argvalues=[False, True])
@pytest.mark.parametrize(argnames='mask_fn', argvalues=[None, lambda x, y: (x ** 2 + y ** 2 < 1)])
@pytest.mark.parametrize(argnames=['nx', 'ny'], argvalues=[(30, 50), (40, 20)])
def test_stream_plot_monitor(mask_fn, nx, ny, specify_field_names):
    nets = [FCNN(2, 1, hidden_units=(3,)) for _ in range(5)]
    conditions = [NoCondition() for _ in nets]
    pairs = [(0, 1), (2, 3), (0, 3), 4, 2]

    if specify_field_names:
        field_names = [str(i) for i in range(len(pairs))]
    else:
        field_names = None
    monitor = StreamPlotMonitor2D(
        xy_min=(-1, -1),
        xy_max=(1, 1),
        nx=nx,
        ny=ny,
        pairs=pairs,
        mask_fn=mask_fn,
        equal_aspect=True,
        field_names=field_names,
    )

    if specify_field_names:
        with pytest.raises(ValueError):
            StreamPlotMonitor2D(
                xy_min=(-1, -1),
                xy_max=(1, 1),
                pairs=[0, 1],
                field_names=['a', 'b', 'c'],
            )

    monitor.check(nets, conditions, history=None)
    monitor.check(nets[::-1], conditions, history=None)



from abc import ABC, abstractmethod
import torch


class NumericalSolver(ABC):
    @abstractmethod
    def solve(self, func, u0, t0, tn, n_steps):
        pass

    @abstractmethod
    def step(self, func, u, t, h):
        pass


class Euler(NumericalSolver):
    order = 1

    def solve(self, func, u0, t0, tn, n_steps, hypernet=None):
        ts = torch.linspace(t0, tn, n_steps + 1)
        if isinstance(u0, (float, int)):
            u0 = (u0,)
        if isinstance(u0, (list, tuple)):
            u0 = torch.tensor(u0)
        us = [u0]
        h = (tn - t0) / n_steps
        for t in ts[:-1]:
            u_old = us[-1]
            u_new = u_old + h * torch.tensor(self.step(func, u_old, t, h))
            if hypernet is not None:
                u_new += h ** 2 * hypernet(torch.cat([t.reshape(1, 1), u_old.reshape(1, -1)], dim=1)).flatten()
            us.append(u_new)

        us = torch.stack(us, dim=0)
        ans = [ts]
        for j in range(us.shape[1]):
            ans.append(us[:, j])

        return ans

    def step(self, func, u, t, h):
        return func(*u, t)



import pytest
import torch
import numpy as np
from neurodiffeq import diff
from neurodiffeq.generators import Generator1D, SamplerGenerator, EnsembleGenerator
from neurodiffeq.networks import FCNN
from neurodiffeq.conditions import NoCondition
from neurodiffeq.operators import cylindrical_curl
from neurodiffeq.operators import cylindrical_grad
from neurodiffeq.operators import cylindrical_div
from neurodiffeq.operators import cylindrical_laplacian
from neurodiffeq.operators import cylindrical_vector_laplacian
from neurodiffeq.operators import cylindrical_to_cartesian, cartesian_to_cylindrical

R_MIN, R_MAX = 1.0, 10.
Z_MIN, Z_MAX = -10., 10.
GRID = (10, 10, 10)


@pytest.fixture
def x():
    g = EnsembleGenerator(
        Generator1D(GRID[0], R_MIN, R_MAX, method='uniform'),
        Generator1D(GRID[1], 0, np.pi * 2, method='uniform'),
        Generator1D(GRID[2], Z_MIN, Z_MAX, method='uniform'),
    )
    return SamplerGenerator(g).get_examples()


@pytest.fixture
def u(x):
    cond = NoCondition()
    net = FCNN(3, 1)
    return cond.enforce(net, *x)


@pytest.fixture
def U(x):
    cond = NoCondition()
    nets = [FCNN(3, 1) for _ in range(3)]
    return tuple(cond.enforce(net, *x) for net in nets)


def test_cylindrical_to_cartesian(x):
    rho, phi, zs = x
    x, y, zc = cylindrical_to_cartesian(*x)

    assert torch.allclose(rho * torch.cos(phi), x)
    assert torch.allclose(rho * torch.sin(phi), y)
    assert torch.allclose(zs, zc)


def test_cartesian_to_cylindrical():
    x = torch.rand(1024, requires_grad=True)
    y = torch.rand(1024, requires_grad=True)
    zc = torch.rand(1024, requires_grad=True)
    rho, phi, zs = cartesian_to_cylindrical(x, y, zc)

    assert torch.allclose(rho * torch.cos(phi), x)
    assert torch.allclose(rho * torch.sin(phi), y)
    assert torch.allclose(zs, zc)


def test_cylindrical_grad(u, x):
    out_rho, out_phi, out_z = cylindrical_grad(u, *x)
    rho, phi, z = x
    assert torch.allclose(out_rho, diff(u, rho))
    assert torch.allclose(out_phi, diff(u, phi) / rho)
    assert torch.allclose(out_z, diff(u, z))


def test_cylindrical_div(U, x):
    out = cylindrical_div(*U, *x)
    rho, phi, z = x
    urho, uphi, uz = U

    assert torch.allclose(out, diff(rho * urho, rho) / rho + diff(uphi, phi) / rho + diff(uz, z))


def test_cylindrical_curl(U, x):
    out_rho, out_phi, out_z = cylindrical_curl(*U, *x)
    rho, phi, z = x
    urho, uphi, uz = U

    assert torch.allclose(out_rho, diff(uz, phi) / rho - diff(uphi, z))
    assert torch.allclose(out_phi, diff(urho, z) - diff(uz, rho))
    assert torch.allclose(out_z, (diff(rho * uphi, rho) - diff(urho, phi)) / rho)


def test_cylindrical_laplacian(u, x):
    out = cylindrical_laplacian(u, *x)
    rho, phi, z = x

    assert torch.allclose(
        out,
        diff(rho * diff(u, rho), rho) / rho + diff(u, phi, order=2) / rho ** 2 + diff(u, z, order=2)
    )


def test_cylindrical_vector_laplacian(U, x):
    out_rho, out_phi, out_z = cylindrical_vector_laplacian(*U, *x)
    rho, phi, z = x
    urho, uphi, uz = U

    def scalar_lap(u):
        return diff(rho * diff(u, rho), rho) / rho + diff(u, phi, order=2) / rho ** 2 + diff(u, z, order=2)

    assert torch.allclose(out_rho, scalar_lap(urho) - urho / rho ** 2 - 2 / rho ** 2 * diff(uphi, phi))
    assert torch.allclose(out_phi, scalar_lap(uphi) - uphi / rho ** 2 + 2 / rho ** 2 * diff(urho, phi))
    assert torch.allclose(out_z, scalar_lap(uz))



from neurodiffeq.hypersolver.numerical_solvers import Euler
import torch
import numpy as np


def test_euler():
    try_solver = Euler()
    ts, us = try_solver.solve(lambda u, t: [-u], 1, 0, 1, 100)
    vs = torch.exp(-ts)
    print(us.shape, vs.shape)
    assert torch.allclose(us, vs, rtol=1e-2)

    ts, fs, gs = try_solver.solve(lambda u, v, t: [v, -u], [1, 1], 0, np.pi, 314)
    fs_true = torch.sin(ts) + torch.cos(ts)
    gs_true = torch.cos(ts) - torch.sin(ts)
    assert torch.allclose(fs, fs_true, rtol=1e-2, atol=1e-2)
    assert torch.allclose(gs, gs_true, rtol=1e-2, atol=1e-2)


import torch
from .operators import grad


def _l1_norm(residual, funcs, coords):
    return torch.abs(residual).mean()


def _l2_norm(residual, funcs, coords):
    return (residual ** 2).mean()


def _infinity_norm(residual, funcs, coords):
    return residual.abs().max(dim=1)[0].mean()


def _h1_norm(residual, funcs, coords):
    g = grad(residual, *coords)
    rg = torch.cat([residual, *g], dim=1)
    return (rg ** 2).mean()


def _h1_semi_norm(residual, funcs, coords):
    g = grad(residual, *coords)
    g = torch.cat(g, dim=1)
    return (g ** 2).mean()


_losses = {
    'l1': _l1_norm,
    'l2': _l2_norm,
    'infinity': _infinity_norm,
    'h1': _h1_norm,
    'h1 semi': _h1_semi_norm,
}
